from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.Helper.regex import (ActivePatterns__007CUnitColumnHeader_007C__007C, ActivePatterns__007CTANColumnHeader_007C__007C, ActivePatterns__007CTSRColumnHeader_007C__007C, try_parse_parameter_column_header, try_parse_factor_column_header, try_parse_characteristic_column_header, try_parse_component_column_header, ActivePatterns__007CInputColumnHeader_007C__007C, ActivePatterns__007COutputColumnHeader_007C__007C, ActivePatterns__007CComment_007C__007C)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Table.composite_cell import CompositeCell
from ...arctrl_core.Table.composite_header import (CompositeHeader, IOType)
from ...fable_library.array_ import (equals_with, skip, try_find_index)
from ...fable_library.option import map
from ...fable_library.seq import (to_array, delay, append, singleton, empty)
from ...fable_library.string_ import (to_fail, printf, starts_with_exact)
from ...fable_library.types import (Array, to_string)
from ...fable_library.util import IEnumerable_1
from .composite_cell import (term_from_string_cells, unitized_from_string_cells, data_from_string_cells, free_text_from_string_cells)

def ActivePattern_mergeIDInfo(id_space1: str, local_id1: str, id_space2: str, local_id2: str) -> dict[str, Any]:
    if id_space1 != id_space2:
        to_fail(printf("TermSourceRef %s and %s do not match"))(id_space1)(id_space2)

    if local_id1 != local_id2:
        to_fail(printf("LocalID %s and %s do not match"))(local_id1)(local_id2)

    return {
        "TermAccessionNumber": ((("" + id_space1) + ":") + local_id1) + "",
        "TermSourceRef": id_space1
    }


def ActivePattern__007CTerm_007C__007C(category_parser: Callable[[str], str | None], f: Callable[[OntologyAnnotation], CompositeHeader], cell_values: Array[str]) -> tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None:
    def _007CAC_007C__007C(s: str, category_parser: Any=category_parser, f: Any=f, cell_values: Any=cell_values) -> str | None:
        return category_parser(s)

    (pattern_matching_result, name, name_1, term1, term2) = (None, None, None, None, None)
    def _arrow3507(x: str, y: str, category_parser: Any=category_parser, f: Any=f, cell_values: Any=cell_values) -> bool:
        return x == y

    if (len(cell_values) == 1) if (not equals_with(_arrow3507, cell_values, None)) else False:
        active_pattern_result: str | None = _007CAC_007C__007C(cell_values[0])
        if active_pattern_result is not None:
            pattern_matching_result = 0
            name = active_pattern_result

        else: 
            pattern_matching_result = 2


    else: 
        def _arrow3508(x_1: str, y_1: str, category_parser: Any=category_parser, f: Any=f, cell_values: Any=cell_values) -> bool:
            return x_1 == y_1

        if (len(cell_values) == 3) if (not equals_with(_arrow3508, cell_values, None)) else False:
            active_pattern_result_1: str | None = _007CAC_007C__007C(cell_values[0])
            if active_pattern_result_1 is not None:
                active_pattern_result_2: dict[str, Any] | None = ActivePatterns__007CTSRColumnHeader_007C__007C(cell_values[1])
                if active_pattern_result_2 is not None:
                    active_pattern_result_3: dict[str, Any] | None = ActivePatterns__007CTANColumnHeader_007C__007C(cell_values[2])
                    if active_pattern_result_3 is not None:
                        pattern_matching_result = 1
                        name_1 = active_pattern_result_1
                        term1 = active_pattern_result_2
                        term2 = active_pattern_result_3

                    else: 
                        pattern_matching_result = 2


                else: 
                    pattern_matching_result = 2


            else: 
                pattern_matching_result = 2


        else: 
            pattern_matching_result = 2


    if pattern_matching_result == 0:
        def _arrow3499(cell_values_1: Array[str], category_parser: Any=category_parser, f: Any=f, cell_values: Any=cell_values) -> CompositeCell:
            return term_from_string_cells(None, None, cell_values_1)

        return (f(OntologyAnnotation.create(name)), _arrow3499)

    elif pattern_matching_result == 1:
        term: dict[str, Any] = ActivePattern_mergeIDInfo(term1["IDSpace"], term1["LocalID"], term2["IDSpace"], term2["LocalID"])
        def _arrow3500(cell_values_2: Array[str], category_parser: Any=category_parser, f: Any=f, cell_values: Any=cell_values) -> CompositeCell:
            return term_from_string_cells(1, 2, cell_values_2)

        return (f(OntologyAnnotation.create(name_1, term["TermSourceRef"], term["TermAccessionNumber"])), _arrow3500)

    elif pattern_matching_result == 2:
        (pattern_matching_result_1, name_2, term1_1, term2_1) = (None, None, None, None)
        def _arrow3506(x_2: str, y_2: str, category_parser: Any=category_parser, f: Any=f, cell_values: Any=cell_values) -> bool:
            return x_2 == y_2

        if (len(cell_values) == 3) if (not equals_with(_arrow3506, cell_values, None)) else False:
            active_pattern_result_4: str | None = _007CAC_007C__007C(cell_values[0])
            if active_pattern_result_4 is not None:
                active_pattern_result_5: dict[str, Any] | None = ActivePatterns__007CTANColumnHeader_007C__007C(cell_values[1])
                if active_pattern_result_5 is not None:
                    active_pattern_result_6: dict[str, Any] | None = ActivePatterns__007CTSRColumnHeader_007C__007C(cell_values[2])
                    if active_pattern_result_6 is not None:
                        pattern_matching_result_1 = 0
                        name_2 = active_pattern_result_4
                        term1_1 = active_pattern_result_6
                        term2_1 = active_pattern_result_5

                    else: 
                        pattern_matching_result_1 = 1


                else: 
                    pattern_matching_result_1 = 1


            else: 
                pattern_matching_result_1 = 1


        else: 
            pattern_matching_result_1 = 1

        if pattern_matching_result_1 == 0:
            term_1: dict[str, Any] = ActivePattern_mergeIDInfo(term1_1["IDSpace"], term1_1["LocalID"], term2_1["IDSpace"], term2_1["LocalID"])
            def _arrow3501(cell_values_3: Array[str], category_parser: Any=category_parser, f: Any=f, cell_values: Any=cell_values) -> CompositeCell:
                return term_from_string_cells(2, 1, cell_values_3)

            return (f(OntologyAnnotation.create(name_2, term_1["TermSourceRef"], term_1["TermAccessionNumber"])), _arrow3501)

        elif pattern_matching_result_1 == 1:
            (pattern_matching_result_2, name_3, term1_2, term2_2) = (None, None, None, None)
            def _arrow3505(x_3: str, y_3: str, category_parser: Any=category_parser, f: Any=f, cell_values: Any=cell_values) -> bool:
                return x_3 == y_3

            if (len(cell_values) == 4) if (not equals_with(_arrow3505, cell_values, None)) else False:
                active_pattern_result_7: str | None = _007CAC_007C__007C(cell_values[0])
                if active_pattern_result_7 is not None:
                    if ActivePatterns__007CUnitColumnHeader_007C__007C(cell_values[1]) is not None:
                        active_pattern_result_9: dict[str, Any] | None = ActivePatterns__007CTSRColumnHeader_007C__007C(cell_values[2])
                        if active_pattern_result_9 is not None:
                            active_pattern_result_10: dict[str, Any] | None = ActivePatterns__007CTANColumnHeader_007C__007C(cell_values[3])
                            if active_pattern_result_10 is not None:
                                pattern_matching_result_2 = 0
                                name_3 = active_pattern_result_7
                                term1_2 = active_pattern_result_9
                                term2_2 = active_pattern_result_10

                            else: 
                                pattern_matching_result_2 = 1


                        else: 
                            pattern_matching_result_2 = 1


                    else: 
                        pattern_matching_result_2 = 1


                else: 
                    pattern_matching_result_2 = 1


            else: 
                pattern_matching_result_2 = 1

            if pattern_matching_result_2 == 0:
                term_2: dict[str, Any] = ActivePattern_mergeIDInfo(term1_2["IDSpace"], term1_2["LocalID"], term2_2["IDSpace"], term2_2["LocalID"])
                def _arrow3502(cell_values_4: Array[str], category_parser: Any=category_parser, f: Any=f, cell_values: Any=cell_values) -> CompositeCell:
                    return unitized_from_string_cells(1, 2, 3, cell_values_4)

                return (f(OntologyAnnotation.create(name_3, term_2["TermSourceRef"], term_2["TermAccessionNumber"])), _arrow3502)

            elif pattern_matching_result_2 == 1:
                (pattern_matching_result_3, name_4, term1_3, term2_3) = (None, None, None, None)
                def _arrow3504(x_4: str, y_4: str, category_parser: Any=category_parser, f: Any=f, cell_values: Any=cell_values) -> bool:
                    return x_4 == y_4

                if (len(cell_values) == 4) if (not equals_with(_arrow3504, cell_values, None)) else False:
                    active_pattern_result_11: str | None = _007CAC_007C__007C(cell_values[0])
                    if active_pattern_result_11 is not None:
                        if ActivePatterns__007CUnitColumnHeader_007C__007C(cell_values[1]) is not None:
                            active_pattern_result_13: dict[str, Any] | None = ActivePatterns__007CTANColumnHeader_007C__007C(cell_values[2])
                            if active_pattern_result_13 is not None:
                                active_pattern_result_14: dict[str, Any] | None = ActivePatterns__007CTSRColumnHeader_007C__007C(cell_values[3])
                                if active_pattern_result_14 is not None:
                                    pattern_matching_result_3 = 0
                                    name_4 = active_pattern_result_11
                                    term1_3 = active_pattern_result_14
                                    term2_3 = active_pattern_result_13

                                else: 
                                    pattern_matching_result_3 = 1


                            else: 
                                pattern_matching_result_3 = 1


                        else: 
                            pattern_matching_result_3 = 1


                    else: 
                        pattern_matching_result_3 = 1


                else: 
                    pattern_matching_result_3 = 1

                if pattern_matching_result_3 == 0:
                    term_3: dict[str, Any] = ActivePattern_mergeIDInfo(term1_3["IDSpace"], term1_3["LocalID"], term2_3["IDSpace"], term2_3["LocalID"])
                    def _arrow3503(cell_values_5: Array[str], category_parser: Any=category_parser, f: Any=f, cell_values: Any=cell_values) -> CompositeCell:
                        return unitized_from_string_cells(1, 3, 2, cell_values_5)

                    return (f(OntologyAnnotation.create(name_4, term_3["TermSourceRef"], term_3["TermAccessionNumber"])), _arrow3503)

                elif pattern_matching_result_3 == 1:
                    return None






def ActivePattern__007CParameter_007C__007C(cell_values: Array[str]) -> tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None:
    def _arrow3509(input: str, cell_values: Any=cell_values) -> str | None:
        return try_parse_parameter_column_header(input)

    def _arrow3510(Item: OntologyAnnotation, cell_values: Any=cell_values) -> CompositeHeader:
        return CompositeHeader(3, Item)

    active_pattern_result: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None = ActivePattern__007CTerm_007C__007C(_arrow3509, _arrow3510, cell_values)
    if active_pattern_result is not None:
        r: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] = active_pattern_result
        return r

    else: 
        return None



def ActivePattern__007CFactor_007C__007C(cell_values: Array[str]) -> tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None:
    def _arrow3511(input: str, cell_values: Any=cell_values) -> str | None:
        return try_parse_factor_column_header(input)

    def _arrow3512(Item: OntologyAnnotation, cell_values: Any=cell_values) -> CompositeHeader:
        return CompositeHeader(2, Item)

    active_pattern_result: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None = ActivePattern__007CTerm_007C__007C(_arrow3511, _arrow3512, cell_values)
    if active_pattern_result is not None:
        r: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] = active_pattern_result
        return r

    else: 
        return None



def ActivePattern__007CCharacteristic_007C__007C(cell_values: Array[str]) -> tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None:
    def _arrow3513(input: str, cell_values: Any=cell_values) -> str | None:
        return try_parse_characteristic_column_header(input)

    def _arrow3514(Item: OntologyAnnotation, cell_values: Any=cell_values) -> CompositeHeader:
        return CompositeHeader(1, Item)

    active_pattern_result: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None = ActivePattern__007CTerm_007C__007C(_arrow3513, _arrow3514, cell_values)
    if active_pattern_result is not None:
        r: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] = active_pattern_result
        return r

    else: 
        return None



def ActivePattern__007CComponent_007C__007C(cell_values: Array[str]) -> tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None:
    def _arrow3515(input: str, cell_values: Any=cell_values) -> str | None:
        return try_parse_component_column_header(input)

    def _arrow3516(Item: OntologyAnnotation, cell_values: Any=cell_values) -> CompositeHeader:
        return CompositeHeader(0, Item)

    active_pattern_result: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None = ActivePattern__007CTerm_007C__007C(_arrow3515, _arrow3516, cell_values)
    if active_pattern_result is not None:
        r: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] = active_pattern_result
        return r

    else: 
        return None



def ActivePattern__007CInput_007C__007C(cell_values: Array[str]) -> tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None:
    if len(cell_values) == 0:
        return None

    else: 
        match_value: str = cell_values[0]
        active_pattern_result: str | None = ActivePatterns__007CInputColumnHeader_007C__007C(match_value)
        if active_pattern_result is not None:
            io_type: str = active_pattern_result
            cols: Array[str] = skip(1, cell_values, None)
            match_value_1: IOType = IOType.of_string(io_type)
            if match_value_1.tag == 2:
                def mapping(y: int, cell_values: Any=cell_values) -> int:
                    return 1 + y

                def predicate(s: str, cell_values: Any=cell_values) -> bool:
                    return starts_with_exact(s, "Data Format")

                format: int | None = map(mapping, try_find_index(predicate, cols))
                def mapping_1(y_1: int, cell_values: Any=cell_values) -> int:
                    return 1 + y_1

                def predicate_1(s_1: str, cell_values: Any=cell_values) -> bool:
                    return starts_with_exact(s_1, "Data Selector Format")

                selector_format: int | None = map(mapping_1, try_find_index(predicate_1, cols))
                def _arrow3517(cell_values_1: Array[str], cell_values: Any=cell_values) -> CompositeCell:
                    return data_from_string_cells(format, selector_format, cell_values_1)

                return (CompositeHeader(11, IOType(2)), _arrow3517)

            else: 
                def _arrow3518(cell_values_2: Array[str], cell_values: Any=cell_values) -> CompositeCell:
                    return free_text_from_string_cells(cell_values_2)

                return (CompositeHeader(11, match_value_1), _arrow3518)


        else: 
            return None




def ActivePattern__007COutput_007C__007C(cell_values: Array[str]) -> tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None:
    if len(cell_values) == 0:
        return None

    else: 
        match_value: str = cell_values[0]
        active_pattern_result: str | None = ActivePatterns__007COutputColumnHeader_007C__007C(match_value)
        if active_pattern_result is not None:
            io_type: str = active_pattern_result
            cols: Array[str] = skip(1, cell_values, None)
            match_value_1: IOType = IOType.of_string(io_type)
            if match_value_1.tag == 2:
                def mapping(y: int, cell_values: Any=cell_values) -> int:
                    return 1 + y

                def predicate(s: str, cell_values: Any=cell_values) -> bool:
                    return starts_with_exact(s, "Data Format")

                format: int | None = map(mapping, try_find_index(predicate, cols))
                def mapping_1(y_1: int, cell_values: Any=cell_values) -> int:
                    return 1 + y_1

                def predicate_1(s_1: str, cell_values: Any=cell_values) -> bool:
                    return starts_with_exact(s_1, "Data Selector Format")

                selector_format: int | None = map(mapping_1, try_find_index(predicate_1, cols))
                def _arrow3519(cell_values_1: Array[str], cell_values: Any=cell_values) -> CompositeCell:
                    return data_from_string_cells(format, selector_format, cell_values_1)

                return (CompositeHeader(12, IOType(2)), _arrow3519)

            else: 
                def _arrow3520(cell_values_2: Array[str], cell_values: Any=cell_values) -> CompositeCell:
                    return free_text_from_string_cells(cell_values_2)

                return (CompositeHeader(12, match_value_1), _arrow3520)


        else: 
            return None




def ActivePattern__007CComment_007C__007C(cell_values: Array[str]) -> tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None:
    (pattern_matching_result, key) = (None, None)
    def _arrow3522(x: str, y: str, cell_values: Any=cell_values) -> bool:
        return x == y

    if (len(cell_values) == 1) if (not equals_with(_arrow3522, cell_values, None)) else False:
        active_pattern_result: str | None = ActivePatterns__007CComment_007C__007C(cell_values[0])
        if active_pattern_result is not None:
            pattern_matching_result = 0
            key = active_pattern_result

        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        def _arrow3521(cell_values_1: Array[str], cell_values: Any=cell_values) -> CompositeCell:
            return free_text_from_string_cells(cell_values_1)

        return (CompositeHeader(14, key), _arrow3521)

    elif pattern_matching_result == 1:
        return None



def ActivePattern__007CProtocolType_007C__007C(cell_values: Array[str]) -> tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None:
    def parser(s: str, cell_values: Any=cell_values) -> str | None:
        if s == "Protocol Type":
            return s

        else: 
            return None


    def header(_arg: OntologyAnnotation, cell_values: Any=cell_values) -> CompositeHeader:
        return CompositeHeader(4)

    active_pattern_result: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None = ActivePattern__007CTerm_007C__007C(parser, header, cell_values)
    if active_pattern_result is not None:
        r: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] = active_pattern_result
        return r

    else: 
        return None



def ActivePattern__007CProtocolHeader_007C__007C(cell_values: Array[str]) -> tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None:
    (pattern_matching_result,) = (None,)
    def _arrow3529(x: str, y: str, cell_values: Any=cell_values) -> bool:
        return x == y

    if (len(cell_values) == 1) if (not equals_with(_arrow3529, cell_values, None)) else False:
        if cell_values[0] == "Protocol REF":
            pattern_matching_result = 0

        elif cell_values[0] == "Protocol Description":
            pattern_matching_result = 1

        elif cell_values[0] == "Protocol Uri":
            pattern_matching_result = 2

        elif cell_values[0] == "Protocol Version":
            pattern_matching_result = 3

        elif cell_values[0] == "Performer":
            pattern_matching_result = 4

        elif cell_values[0] == "Date":
            pattern_matching_result = 5

        else: 
            pattern_matching_result = 6


    else: 
        pattern_matching_result = 6

    if pattern_matching_result == 0:
        def _arrow3523(cell_values_1: Array[str], cell_values: Any=cell_values) -> CompositeCell:
            return free_text_from_string_cells(cell_values_1)

        return (CompositeHeader(8), _arrow3523)

    elif pattern_matching_result == 1:
        def _arrow3524(cell_values_2: Array[str], cell_values: Any=cell_values) -> CompositeCell:
            return free_text_from_string_cells(cell_values_2)

        return (CompositeHeader(5), _arrow3524)

    elif pattern_matching_result == 2:
        def _arrow3525(cell_values_3: Array[str], cell_values: Any=cell_values) -> CompositeCell:
            return free_text_from_string_cells(cell_values_3)

        return (CompositeHeader(6), _arrow3525)

    elif pattern_matching_result == 3:
        def _arrow3526(cell_values_4: Array[str], cell_values: Any=cell_values) -> CompositeCell:
            return free_text_from_string_cells(cell_values_4)

        return (CompositeHeader(7), _arrow3526)

    elif pattern_matching_result == 4:
        def _arrow3527(cell_values_5: Array[str], cell_values: Any=cell_values) -> CompositeCell:
            return free_text_from_string_cells(cell_values_5)

        return (CompositeHeader(9), _arrow3527)

    elif pattern_matching_result == 5:
        def _arrow3528(cell_values_6: Array[str], cell_values: Any=cell_values) -> CompositeCell:
            return free_text_from_string_cells(cell_values_6)

        return (CompositeHeader(10), _arrow3528)

    elif pattern_matching_result == 6:
        return None



def ActivePattern__007CFreeText_007C__007C(cell_values: Array[str]) -> tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None:
    def _arrow3530(x: str, y: str, cell_values: Any=cell_values) -> bool:
        return x == y

    if (len(cell_values) == 1) if (not equals_with(_arrow3530, cell_values, None)) else False:
        def _arrow3531(cell_values_1: Array[str], cell_values: Any=cell_values) -> CompositeCell:
            return free_text_from_string_cells(cell_values_1)

        return (CompositeHeader(13, cell_values[0]), _arrow3531)

    else: 
        return None



def from_string_cells(cell_values: Array[str]) -> tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]]:
    active_pattern_result: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None = ActivePattern__007CParameter_007C__007C(cell_values)
    if active_pattern_result is not None:
        p: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] = active_pattern_result
        return p

    else: 
        active_pattern_result_1: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None = ActivePattern__007CFactor_007C__007C(cell_values)
        if active_pattern_result_1 is not None:
            f: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] = active_pattern_result_1
            return f

        else: 
            active_pattern_result_2: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None = ActivePattern__007CCharacteristic_007C__007C(cell_values)
            if active_pattern_result_2 is not None:
                c: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] = active_pattern_result_2
                return c

            else: 
                active_pattern_result_3: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None = ActivePattern__007CComponent_007C__007C(cell_values)
                if active_pattern_result_3 is not None:
                    c_1: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] = active_pattern_result_3
                    return c_1

                else: 
                    active_pattern_result_4: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None = ActivePattern__007CInput_007C__007C(cell_values)
                    if active_pattern_result_4 is not None:
                        i: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] = active_pattern_result_4
                        return i

                    else: 
                        active_pattern_result_5: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None = ActivePattern__007COutput_007C__007C(cell_values)
                        if active_pattern_result_5 is not None:
                            o: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] = active_pattern_result_5
                            return o

                        else: 
                            active_pattern_result_6: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None = ActivePattern__007CProtocolType_007C__007C(cell_values)
                            if active_pattern_result_6 is not None:
                                pt: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] = active_pattern_result_6
                                return pt

                            else: 
                                active_pattern_result_7: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None = ActivePattern__007CProtocolHeader_007C__007C(cell_values)
                                if active_pattern_result_7 is not None:
                                    ph: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] = active_pattern_result_7
                                    return ph

                                else: 
                                    active_pattern_result_8: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None = ActivePattern__007CComment_007C__007C(cell_values)
                                    if active_pattern_result_8 is not None:
                                        c_2: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] = active_pattern_result_8
                                        return c_2

                                    else: 
                                        active_pattern_result_9: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] | None = ActivePattern__007CFreeText_007C__007C(cell_values)
                                        if active_pattern_result_9 is not None:
                                            ft: tuple[CompositeHeader, Callable[[Array[str]], CompositeCell]] = active_pattern_result_9
                                            return ft

                                        else: 
                                            return to_fail(printf("Could not parse header group %O"))(cell_values)












def to_string_cells(has_unit: bool, header: CompositeHeader) -> Array[str]:
    if header.IsDataColumn:
        return [to_string(header), "Data Format", "Data Selector Format"]

    elif header.IsSingleColumn:
        return [to_string(header)]

    elif header.IsTermColumn:
        def _arrow3535(__unit: None=None, has_unit: Any=has_unit, header: Any=header) -> IEnumerable_1[str]:
            def _arrow3534(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow3533(__unit: None=None) -> IEnumerable_1[str]:
                    def _arrow3532(__unit: None=None) -> IEnumerable_1[str]:
                        return singleton(("Term Accession Number (" + header.GetColumnAccessionShort) + ")")

                    return append(singleton(("Term Source REF (" + header.GetColumnAccessionShort) + ")"), delay(_arrow3532))

                return append(singleton("Unit") if has_unit else empty(), delay(_arrow3533))

            return append(singleton(to_string(header)), delay(_arrow3534))

        return to_array(delay(_arrow3535))

    else: 
        return to_fail(printf("header %O is neither single nor term column"))(header)



__all__ = ["ActivePattern_mergeIDInfo", "ActivePattern__007CTerm_007C__007C", "ActivePattern__007CParameter_007C__007C", "ActivePattern__007CFactor_007C__007C", "ActivePattern__007CCharacteristic_007C__007C", "ActivePattern__007CComponent_007C__007C", "ActivePattern__007CInput_007C__007C", "ActivePattern__007COutput_007C__007C", "ActivePattern__007CComment_007C__007C", "ActivePattern__007CProtocolType_007C__007C", "ActivePattern__007CProtocolHeader_007C__007C", "ActivePattern__007CFreeText_007C__007C", "from_string_cells", "to_string_cells"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...fable_library.array_ import contains
from ...fable_library.option import value
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import Array
from ...fable_library.util import string_hash
from ..helper import clean
from ..ldcontext import LDContext
from ..ldobject import (LDNode, LDGraph)
from .property_value import LDPropertyValue

def _expr1514() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDSample", None, LDSample)


class LDSample:
    @staticmethod
    def schema_type() -> str:
        return "https://bioschemas.org/Sample"

    @staticmethod
    def name() -> str:
        return "http://schema.org/name"

    @staticmethod
    def additional_property() -> str:
        return "http://schema.org/additionalProperty"

    @staticmethod
    def try_get_name_as_string(s: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = s.TryGetPropertyAsSingleton(LDSample.name(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_name_as_string(s: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = s.TryGetPropertyAsSingleton(LDSample.name(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            raise Exception(("Could not access property `name` of object with @id `" + s.Id) + "`")


    @staticmethod
    def set_name_as_string(s: LDNode, n: str) -> None:
        s.SetProperty(LDSample.name(), n)

    @staticmethod
    def get_additional_properties(s: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDPropertyValue.validate(ld_object, context_1)

        return s.GetPropertyNodes(LDSample.additional_property(), filter, graph, context)

    @staticmethod
    def set_additional_properties(s: LDNode, additional_properties: Array[LDNode], context: LDContext | None=None) -> Any:
        return s.SetProperty(LDSample.additional_property(), additional_properties, context)

    @staticmethod
    def get_characteristics(s: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDPropertyValue.validate_characteristic_value(ld_object, context_1)

        return s.GetPropertyNodes(LDSample.additional_property(), filter, graph, context)

    @staticmethod
    def get_factors(s: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDPropertyValue.validate_factor_value(ld_object, context_1)

        return s.GetPropertyNodes(LDSample.additional_property(), filter, graph, context)

    @staticmethod
    def validate(s: LDNode, context: LDContext | None=None) -> bool:
        return s.HasProperty(LDSample.name(), context) if s.HasType(LDSample.schema_type(), context) else False

    @staticmethod
    def gen_idsample(name: str) -> str:
        return clean(("#Sample_" + name) + "")

    @staticmethod
    def gen_idsource(name: str) -> str:
        return clean(("#Source_" + name) + "")

    @staticmethod
    def gen_idmaterial(name: str) -> str:
        return clean(("#Material_" + name) + "")

    @staticmethod
    def validate_sample(s: LDNode, context: LDContext | None=None) -> bool:
        class ObjectExpr1509:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow1508(x: str, y: str) -> bool:
                    return x == y

                return _arrow1508

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        return contains("Sample", s.AdditionalType, ObjectExpr1509()) if LDSample.validate(s, context) else False

    @staticmethod
    def validate_source(s: LDNode, context: LDContext | None=None) -> bool:
        class ObjectExpr1511:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow1510(x: str, y: str) -> bool:
                    return x == y

                return _arrow1510

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        return contains("Source", s.AdditionalType, ObjectExpr1511()) if LDSample.validate(s, context) else False

    @staticmethod
    def validate_material(s: LDNode, context: LDContext | None=None) -> bool:
        class ObjectExpr1513:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow1512(x: str, y: str) -> bool:
                    return x == y

                return _arrow1512

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        return contains("Material", s.AdditionalType, ObjectExpr1513()) if LDSample.validate(s, context) else False

    @staticmethod
    def create(id: str, name: str, additional_properties: Array[LDNode] | None=None, context: LDContext | None=None) -> LDNode:
        s: LDNode = LDNode(id, [LDSample.schema_type()], None, context)
        s.SetProperty(LDSample.name(), name, context)
        s.SetOptionalProperty(LDSample.additional_property(), additional_properties, context)
        return s

    @staticmethod
    def create_sample(name: str, id: str | None=None, additional_properties: Array[LDNode] | None=None, context: LDContext | None=None) -> LDNode:
        id_2: str = LDSample.gen_idsample(name) if (id is None) else id
        s: LDNode = LDSample.create(id_2, name, additional_properties, context)
        s.AdditionalType = ["Sample"]
        return s

    @staticmethod
    def create_source(name: str, id: str | None=None, additional_properties: Array[LDNode] | None=None, context: LDContext | None=None) -> LDNode:
        id_2: str = LDSample.gen_idsource(name) if (id is None) else id
        s: LDNode = LDSample.create(id_2, name, additional_properties, context)
        s.AdditionalType = ["Source"]
        return s

    @staticmethod
    def create_material(name: str, id: str | None=None, additional_properties: Array[LDNode] | None=None, context: LDContext | None=None) -> LDNode:
        id_2: str = LDSample.gen_idmaterial(name) if (id is None) else id
        s: LDNode = LDSample.create(id_2, name, additional_properties, context)
        s.AdditionalType = ["Material"]
        return s


LDSample_reflection = _expr1514

__all__ = ["LDSample_reflection"]




from __future__ import annotations
from typing import Any
from ...arctrl_core.Helper.collections_ import ResizeArray_map
from ...arctrl_core.Helper.orcid import try_get_orcid_url
from ...fable_library.option import (value as value_1, map, some, bind)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import Array
from ..helper import clean
from ..ldcontext import LDContext
from ..ldobject import (LDNode, LDGraph, LDRef)
from .defined_term import LDDefinedTerm
from .organization import LDOrganization
from .postal_address import LDPostalAddress

def _expr1505() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDPerson", None, LDPerson)


class LDPerson:
    @staticmethod
    def schema_type() -> str:
        return "http://schema.org/Person"

    @staticmethod
    def given_name() -> str:
        return "http://schema.org/givenName"

    @staticmethod
    def affiliation() -> str:
        return "http://schema.org/affiliation"

    @staticmethod
    def email() -> str:
        return "http://schema.org/email"

    @staticmethod
    def family_name() -> str:
        return "http://schema.org/familyName"

    @staticmethod
    def identifier() -> str:
        return "http://schema.org/identifier"

    @staticmethod
    def job_title() -> str:
        return "http://schema.org/jobTitle"

    @staticmethod
    def additional_name() -> str:
        return "http://schema.org/additionalName"

    @staticmethod
    def address() -> str:
        return "http://schema.org/address"

    @staticmethod
    def disambiguating_description() -> str:
        return "http://schema.org/disambiguatingDescription"

    @staticmethod
    def fax_number() -> str:
        return "http://schema.org/faxNumber"

    @staticmethod
    def telephone() -> str:
        return "http://schema.org/telephone"

    @staticmethod
    def try_get_given_name_as_string(p: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = p.TryGetPropertyAsSingleton(LDPerson.given_name(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_given_name_as_string(p: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = p.TryGetPropertyAsSingleton(LDPerson.given_name(), context)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                n: str = value_1(match_value)
                return n

            else: 
                raise Exception(("Property of `givenName` of object with @id `" + p.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `givenName` of object with @id `" + p.Id) + "`")


    @staticmethod
    def set_given_name_as_string(p: LDNode, n: str, context: LDContext | None=None) -> Any:
        return p.SetProperty(LDPerson.given_name(), n, context)

    @staticmethod
    def try_get_affiliation(p: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode | None:
        match_value: LDNode | None = p.TryGetPropertyAsSingleNode(LDPerson.affiliation(), graph, context)
        (pattern_matching_result, n_1) = (None, None)
        if match_value is not None:
            def _arrow1501(__unit: None=None) -> bool:
                n: LDNode = match_value
                return LDOrganization.validate(n, context)

            if _arrow1501():
                pattern_matching_result = 0
                n_1 = match_value

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n_1

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_affiliation(p: LDNode, a: LDNode, context: LDContext | None=None) -> Any:
        return p.SetProperty(LDPerson.affiliation(), a, context)

    @staticmethod
    def try_get_email_as_string(p: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = p.TryGetPropertyAsSingleton(LDPerson.email(), context)
        (pattern_matching_result, e) = (None, None)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                e = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return e

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_email_as_string(p: LDNode, e: str, context: LDContext | None=None) -> Any:
        return p.SetProperty(LDPerson.email(), e, context)

    @staticmethod
    def try_get_family_name_as_string(p: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = p.TryGetPropertyAsSingleton(LDPerson.family_name(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_family_name_as_string(p: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = p.TryGetPropertyAsSingleton(LDPerson.family_name(), context)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                n: str = value_1(match_value)
                return n

            else: 
                raise Exception(("Property of `familyName` of object with @id `" + p.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `familyName` of object with @id `" + p.Id) + "`")


    @staticmethod
    def set_family_name_as_string(p: LDNode, n: str, context: LDContext | None=None) -> Any:
        return p.SetProperty(LDPerson.family_name(), n, context)

    @staticmethod
    def try_get_identifier(p: LDNode, context: LDContext | None=None) -> LDNode | None:
        match_value: Any | None = p.TryGetPropertyAsSingleton(LDPerson.identifier(), context)
        (pattern_matching_result, i) = (None, None)
        if match_value is not None:
            if isinstance(value_1(match_value), LDNode):
                pattern_matching_result = 0
                i = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return i

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_identifier(p: LDNode, i: LDNode, context: LDContext | None=None) -> Any:
        return p.SetProperty(LDPerson.identifier(), i, context)

    @staticmethod
    def get_job_titles_as_defined_term(p: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDDefinedTerm.validate(ld_object, context_1)

        return p.GetPropertyNodes(LDPerson.job_title(), filter, graph, context)

    @staticmethod
    def set_job_title_as_defined_term(p: LDNode, j: Array[LDNode], context: LDContext | None=None) -> Any:
        return p.SetProperty(LDPerson.job_title(), j, context)

    @staticmethod
    def try_get_additional_name_as_string(p: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = p.TryGetPropertyAsSingleton(LDPerson.additional_name(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_additional_name_as_string(p: LDNode, n: str, context: LDContext | None=None) -> Any:
        return p.SetProperty(LDPerson.additional_name(), n, context)

    @staticmethod
    def try_get_address(p: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Any | None:
        match_value: Any | None = p.TryGetPropertyAsSingleton(LDPerson.address(), context)
        (pattern_matching_result, r_1, a, s) = (None, None, None, None)
        if match_value is not None:
            if isinstance(value_1(match_value), LDRef):
                def _arrow1502(__unit: None=None) -> bool:
                    r: LDRef = value_1(match_value)
                    return graph is not None

                if _arrow1502():
                    pattern_matching_result = 0
                    r_1 = value_1(match_value)

                else: 
                    pattern_matching_result = 3


            elif isinstance(value_1(match_value), LDNode):
                pattern_matching_result = 1
                a = value_1(match_value)

            elif str(type(value_1(match_value))) == "<class \'str\'>":
                pattern_matching_result = 2
                s = value_1(match_value)

            else: 
                pattern_matching_result = 3


        else: 
            pattern_matching_result = 3

        if pattern_matching_result == 0:
            def mapping(value: LDNode) -> Any:
                return value

            return map(mapping, value_1(graph).TryGetNode(r_1.Id))

        elif pattern_matching_result == 1:
            return some(a)

        elif pattern_matching_result == 2:
            return some(s)

        elif pattern_matching_result == 3:
            return None


    @staticmethod
    def try_get_address_as_postal_address(p: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode | None:
        match_value: LDNode | None = p.TryGetPropertyAsSingleNode(LDPerson.address(), graph, context)
        (pattern_matching_result, n_1) = (None, None)
        if match_value is not None:
            def _arrow1503(__unit: None=None) -> bool:
                n: LDNode = match_value
                return LDPostalAddress.validate(n, context)

            if _arrow1503():
                pattern_matching_result = 0
                n_1 = match_value

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n_1

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def try_get_address_as_string(p: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = p.TryGetPropertyAsSingleton(LDPerson.address(), context)
        (pattern_matching_result, a) = (None, None)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                a = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return a

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_address_as_postal_address(p: LDNode, a: LDNode, context: LDContext | None=None) -> Any:
        return p.SetProperty(LDPerson.address(), a, context)

    @staticmethod
    def set_address_as_string(p: LDNode, a: str, context: LDContext | None=None) -> Any:
        return p.SetProperty(LDPerson.address(), a, context)

    @staticmethod
    def get_disambiguating_descriptions_as_string(p: LDNode, context: LDContext | None=None) -> Array[str]:
        def f(v: Any=None) -> Any:
            return v

        def filter(value: Any=None, context_1: LDContext | None=None) -> bool:
            return str(type(value)) == "<class \'str\'>"

        return ResizeArray_map(f, p.GetPropertyValues(LDPerson.disambiguating_description(), filter, context))

    @staticmethod
    def set_disambiguating_descriptions_as_string(p: LDNode, d: Array[str], context: LDContext | None=None) -> Any:
        return p.SetProperty(LDPerson.disambiguating_description(), d, context)

    @staticmethod
    def try_get_fax_number_as_string(p: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = p.TryGetPropertyAsSingleton(LDPerson.fax_number(), context)
        (pattern_matching_result, f) = (None, None)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                f = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return f

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_fax_number_as_string(p: LDNode, f: str, context: LDContext | None=None) -> Any:
        return p.SetProperty(LDPerson.fax_number(), f, context)

    @staticmethod
    def try_get_telephone_as_string(p: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = p.TryGetPropertyAsSingleton(LDPerson.telephone(), context)
        (pattern_matching_result, t) = (None, None)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                t = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return t

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_telephone_as_string(p: LDNode, t: str, context: LDContext | None=None) -> Any:
        return p.SetProperty(LDPerson.telephone(), t, context)

    @staticmethod
    def gen_id(given_name: Any, orcid: str | None=None, family_name: Any | None=None) -> str:
        def _arrow1504(__unit: None=None) -> str:
            match_value: str | None = bind(try_get_orcid_url, orcid)
            return ((("#Person_" + str(given_name)) + "") if (family_name is None) else (((("#Person_" + str(given_name)) + "_") + str(value_1(family_name))) + "")) if (match_value is None) else match_value

        return clean(_arrow1504())

    @staticmethod
    def validate(p: LDNode, context: LDContext | None=None) -> bool:
        return p.HasProperty(LDPerson.given_name(), context) if p.HasType(LDPerson.schema_type(), context) else False

    @staticmethod
    def create(given_name: str, orcid: str | None=None, id: str | None=None, affiliation: Any | None=None, email: str | None=None, family_name: str | None=None, identifier: Any | None=None, job_titles: Array[LDNode] | None=None, additional_name: str | None=None, address: Any | None=None, disambiguating_descriptions: Array[str] | None=None, fax_number: str | None=None, telephone: str | None=None, context: LDContext | None=None) -> LDNode:
        person: LDNode = LDNode(LDPerson.gen_id(given_name, orcid, family_name) if (id is None) else id, [LDPerson.schema_type()], None, context)
        person.SetProperty(LDPerson.given_name(), given_name, context)
        person.SetOptionalProperty(LDPerson.affiliation(), affiliation, context)
        person.SetOptionalProperty(LDPerson.email(), email, context)
        person.SetOptionalProperty(LDPerson.family_name(), family_name, context)
        person.SetOptionalProperty(LDPerson.identifier(), identifier, context)
        person.SetOptionalProperty(LDPerson.job_title(), job_titles, context)
        person.SetOptionalProperty(LDPerson.additional_name(), additional_name, context)
        person.SetOptionalProperty(LDPerson.address(), address, context)
        person.SetOptionalProperty(LDPerson.disambiguating_description(), disambiguating_descriptions, context)
        person.SetOptionalProperty(LDPerson.fax_number(), fax_number, context)
        person.SetOptionalProperty(LDPerson.telephone(), telephone, context)
        return person


LDPerson_reflection = _expr1505

__all__ = ["LDPerson_reflection"]




from __future__ import annotations
from typing import (Any, TypeVar)
from ...fable_library.seq import map
from ...fable_library.util import (to_enumerable, IEnumerable_1)
from ...thoth_json_core.decode import (object, IRequiredGetter, string, IGetters)
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from ..Ranges.fs_range_address import FsRangeAddress
from ..Tables.fs_table import FsTable

__A_ = TypeVar("__A_")

def encode(sheet: FsTable) -> IEncodable:
    def _arrow355(__unit: None=None, sheet: Any=sheet) -> IEncodable:
        value: str = sheet.Name
        class ObjectExpr354(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr354()

    def _arrow357(__unit: None=None, sheet: Any=sheet) -> IEncodable:
        value_1: str = sheet.RangeAddress.Range
        class ObjectExpr356(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_1)

        return ObjectExpr356()

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("name", _arrow355()), ("range", _arrow357())])
    class ObjectExpr358(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any], sheet: Any=sheet) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_2))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_2.encode_object(arg)

    return ObjectExpr358()


def _arrow359(builder: IGetters) -> FsTable:
    n: str
    object_arg: IRequiredGetter = builder.Required
    n = object_arg.Field("name", string)
    r: str
    object_arg_1: IRequiredGetter = builder.Required
    r = object_arg_1.Field("range", string)
    return FsTable(n, FsRangeAddress.from_string(r))


decode: Decoder_1[FsTable] = object(_arrow359)

__all__ = ["encode", "decode"]




from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Generic, TypeVar

from .array_ import chunk_by_size as chunk_by_size_1
from .array_ import fill, sort_in_place_with
from .array_ import fold_back as fold_back_1
from .array_ import fold_back2 as fold_back2_1
from .array_ import iterate as iterate_1
from .array_ import map as map_1
from .array_ import pairwise as pairwise_1
from .array_ import permute as permute_1
from .array_ import scan_back as scan_back_1
from .array_ import split_into as split_into_1
from .array_ import transpose as transpose_1
from .array_ import try_find_back as try_find_back_1
from .array_ import try_find_index_back as try_find_index_back_1
from .array_ import windowed as windowed_1
from .global_ import (
    IGenericAdder_1,
    IGenericAverager_1,
    SR_differentLengths,
    SR_indexOutOfBounds,
    SR_inputMustBeNonNegative,
    SR_inputSequenceEmpty,
    SR_inputSequenceTooLong,
    SR_inputWasEmpty,
    SR_keyNotFoundAlt,
    SR_notEnoughElements,
)
from .option import default_arg, some
from .option import value as value_1
from .reflection import TypeInfo, class_type, option_type, record_type
from .string_ import join
from .types import Array, Record
from .util import (
    IComparer_1,
    IDisposable,
    IEnumerable_1,
    IEnumerator,
    IEqualityComparer_1,
    compare,
    equals,
    get_enumerator,
    ignore,
    is_array_like,
    structural_hash,
    to_iterator,
)


_T = TypeVar("_T")

_STATE = TypeVar("_STATE")

_T1 = TypeVar("_T1")

_T2 = TypeVar("_T2")

__A = TypeVar("__A")

__B = TypeVar("__B")

_U = TypeVar("_U")

_T3 = TypeVar("_T3")

_RESULT = TypeVar("_RESULT")

__C = TypeVar("__C")


def _expr66(gen0: TypeInfo) -> TypeInfo:
    return record_type(
        "ListModule.FSharpList",
        [gen0],
        FSharpList,
        lambda: [("head", gen0), ("tail", option_type(FSharpList_reflection(gen0)))],
    )


@dataclass(eq=False, repr=False, slots=True)
class FSharpList(Record, Generic[_T]):
    head: _T
    tail: FSharpList[_T] | None

    def __str__(self, __unit: None = None) -> str:
        xs: FSharpList[_T] = self
        return ("[" + join("; ", xs)) + "]"

    def __eq__(self, other: Any = None) -> bool:
        xs: FSharpList[_T] = self
        if xs is other:
            return True

        else:

            def loop(xs_1_mut: FSharpList[_T], ys_1_mut: FSharpList[_T]) -> bool:
                while True:
                    (xs_1, ys_1) = (xs_1_mut, ys_1_mut)
                    match_value: FSharpList[_T] | None = xs_1.tail
                    match_value_1: FSharpList[_T] | None = ys_1.tail
                    if match_value is not None:
                        if match_value_1 is not None:
                            xt: FSharpList[_T] = match_value
                            yt: FSharpList[_T] = match_value_1
                            if equals(xs_1.head, ys_1.head):
                                xs_1_mut = xt
                                ys_1_mut = yt
                                continue

                            else:
                                return False

                        else:
                            return False

                    elif match_value_1 is not None:
                        return False

                    else:
                        return True

                    break

            return loop(xs, other)

    def __hash__(self, __unit: None = None) -> int:
        xs: FSharpList[_T] = self

        def loop(i_mut: int, h_mut: int, xs_1_mut: FSharpList[_T]) -> int:
            while True:
                (i, h, xs_1) = (i_mut, h_mut, xs_1_mut)
                match_value: FSharpList[_T] | None = xs_1.tail
                if match_value is not None:
                    t: FSharpList[_T] = match_value
                    if i > 18:
                        return h

                    else:
                        i_mut = i + 1
                        h_mut = ((h << 1) + structural_hash(xs_1.head)) + (631 * i)
                        xs_1_mut = t
                        continue

                else:
                    return h

                break

        return loop(0, 0, xs)

    def to_json(self, __unit: None = None) -> Any:
        this: FSharpList[_T] = self
        return list(this)

    def __cmp__(self, other: Any = None) -> int:
        xs: FSharpList[_T] = self

        def loop(xs_1_mut: FSharpList[_T], ys_1_mut: FSharpList[_T]) -> int:
            while True:
                (xs_1, ys_1) = (xs_1_mut, ys_1_mut)
                match_value: FSharpList[_T] | None = xs_1.tail
                match_value_1: FSharpList[_T] | None = ys_1.tail
                if match_value is not None:
                    if match_value_1 is not None:
                        xt: FSharpList[_T] = match_value
                        yt: FSharpList[_T] = match_value_1
                        c: int = compare(xs_1.head, ys_1.head) or 0
                        if c == 0:
                            xs_1_mut = xt
                            ys_1_mut = yt
                            continue

                        else:
                            return c

                    else:
                        return 1

                elif match_value_1 is not None:
                    return -1

                else:
                    return 0

                break

        return loop(xs, other)

    def GetEnumerator(self, __unit: None = None) -> IEnumerator[_T]:
        xs: FSharpList[_T] = self
        return ListEnumerator_1__ctor_3002E699(xs)

    def __iter__(self) -> IEnumerator[_T]:
        return to_iterator(self.GetEnumerator())

    def System_Collections_IEnumerable_GetEnumerator(self, __unit: None = None) -> IEnumerator[Any]:
        xs: FSharpList[_T] = self
        return get_enumerator(xs)


FSharpList_reflection = _expr66


def _expr67(gen0: TypeInfo) -> TypeInfo:
    return class_type("ListModule.ListEnumerator`1", [gen0], ListEnumerator_1)


class ListEnumerator_1(IDisposable, Generic[_T]):
    def __init__(self, xs: FSharpList[Any]) -> None:
        self.xs: FSharpList[_T] = xs
        self.it: FSharpList[_T] = self.xs
        self.current: _T = None

    def System_Collections_Generic_IEnumerator_1_get_Current(self, __unit: None = None) -> _T:
        _: ListEnumerator_1[_T] = self
        return _.current

    def System_Collections_IEnumerator_get_Current(self, __unit: None = None) -> Any:
        _: ListEnumerator_1[_T] = self
        return _.current

    def System_Collections_IEnumerator_MoveNext(self, __unit: None = None) -> bool:
        _: ListEnumerator_1[_T] = self
        match_value: FSharpList[_T] | None = _.it.tail
        if match_value is not None:
            t: FSharpList[_T] = match_value
            _.current = _.it.head
            _.it = t
            return True

        else:
            return False

    def System_Collections_IEnumerator_Reset(self, __unit: None = None) -> None:
        _: ListEnumerator_1[_T] = self
        _.it = _.xs
        _.current = None

    def Dispose(self, __unit: None = None) -> None:
        pass


ListEnumerator_1_reflection = _expr67


def ListEnumerator_1__ctor_3002E699(xs: FSharpList[Any]) -> ListEnumerator_1[_T]:
    return ListEnumerator_1(xs)


def FSharpList_get_Empty(__unit: None = None) -> FSharpList[Any]:
    return FSharpList(None, None)


def FSharpList_Cons_305B8EAC(x: Any, xs: FSharpList[Any]) -> FSharpList[Any]:
    return FSharpList(x, xs)


def FSharpList__get_IsEmpty(xs: FSharpList[Any]) -> bool:
    return xs.tail is None


def FSharpList__get_Length(xs: FSharpList[Any]) -> int:
    def loop(i_mut: int, xs_1_mut: FSharpList[_T], xs: Any = xs) -> int:
        while True:
            (i, xs_1) = (i_mut, xs_1_mut)
            match_value: FSharpList[_T] | None = xs_1.tail
            if match_value is not None:
                i_mut = i + 1
                xs_1_mut = match_value
                continue

            else:
                return i

            break

    return loop(0, xs)


def FSharpList__get_Head(xs: FSharpList[Any]) -> Any:
    match_value: FSharpList[_T] | None = xs.tail
    if match_value is not None:
        return xs.head

    else:
        raise Exception((SR_inputWasEmpty + "\\nParameter name: ") + "list")


def FSharpList__get_Tail(xs: FSharpList[Any]) -> FSharpList[Any]:
    match_value: FSharpList[_T] | None = xs.tail
    if match_value is not None:
        return match_value

    else:
        raise Exception((SR_inputWasEmpty + "\\nParameter name: ") + "list")


def FSharpList__get_Item_Z524259A4(xs: FSharpList[Any], index: int) -> Any:
    def loop(i_mut: int, xs_1_mut: FSharpList[_T], xs: Any = xs, index: Any = index) -> _T:
        while True:
            (i, xs_1) = (i_mut, xs_1_mut)
            match_value: FSharpList[_T] | None = xs_1.tail
            if match_value is not None:
                if i == index:
                    return xs_1.head

                else:
                    i_mut = i + 1
                    xs_1_mut = match_value
                    continue

            else:
                raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")

            break

    return loop(0, xs)


def index_not_found(__unit: None = None) -> Any:
    raise Exception(SR_keyNotFoundAlt)


def empty(__unit: None = None) -> FSharpList[Any]:
    return FSharpList_get_Empty()


def cons(x: Any, xs: FSharpList[Any]) -> FSharpList[Any]:
    return FSharpList_Cons_305B8EAC(x, xs)


def singleton(x: Any | None = None) -> FSharpList[Any]:
    return FSharpList_Cons_305B8EAC(x, FSharpList_get_Empty())


def is_empty(xs: FSharpList[Any]) -> bool:
    return FSharpList__get_IsEmpty(xs)


def length(xs: FSharpList[Any]) -> int:
    return FSharpList__get_Length(xs)


def head(xs: FSharpList[Any]) -> Any:
    return FSharpList__get_Head(xs)


def try_head(xs: FSharpList[Any]) -> Any | None:
    if FSharpList__get_IsEmpty(xs):
        return None

    else:
        return some(FSharpList__get_Head(xs))


def tail(xs: FSharpList[Any]) -> FSharpList[Any]:
    return FSharpList__get_Tail(xs)


def try_last(xs_mut: FSharpList[Any]) -> Any | None:
    while True:
        (xs,) = (xs_mut,)
        if FSharpList__get_IsEmpty(xs):
            return None

        else:
            t: FSharpList[_T] = FSharpList__get_Tail(xs)
            if FSharpList__get_IsEmpty(t):
                return some(FSharpList__get_Head(xs))

            else:
                xs_mut = t
                continue

        break


def last(xs: FSharpList[Any]) -> Any:
    match_value: _T | None = try_last(xs)
    if match_value is None:
        raise Exception(SR_inputWasEmpty)

    else:
        return value_1(match_value)


def compare_with(comparer: Callable[[_T, _T], int], xs: FSharpList[Any], ys: FSharpList[Any]) -> int:
    def loop(
        xs_1_mut: FSharpList[_T], ys_1_mut: FSharpList[_T], comparer: Any = comparer, xs: Any = xs, ys: Any = ys
    ) -> int:
        while True:
            (xs_1, ys_1) = (xs_1_mut, ys_1_mut)
            match_value: bool = FSharpList__get_IsEmpty(xs_1)
            match_value_1: bool = FSharpList__get_IsEmpty(ys_1)
            if match_value:
                if match_value_1:
                    return 0

                else:
                    return -1

            elif match_value_1:
                return 1

            else:
                c: int = comparer(FSharpList__get_Head(xs_1), FSharpList__get_Head(ys_1)) or 0
                if c == 0:
                    xs_1_mut = FSharpList__get_Tail(xs_1)
                    ys_1_mut = FSharpList__get_Tail(ys_1)
                    continue

                else:
                    return c

            break

    return loop(xs, ys)


def to_array(xs: FSharpList[Any]) -> Array[Any]:
    len_1: int = FSharpList__get_Length(xs) or 0
    res: Array[_T] = fill([0] * len_1, 0, len_1, None)

    def loop(i_mut: int, xs_1_mut: FSharpList[_T], xs: Any = xs) -> None:
        while True:
            (i, xs_1) = (i_mut, xs_1_mut)
            if not FSharpList__get_IsEmpty(xs_1):
                res[i] = FSharpList__get_Head(xs_1)
                i_mut = i + 1
                xs_1_mut = FSharpList__get_Tail(xs_1)
                continue

            break

    loop(0, xs)
    return res


def fold(folder: Callable[[_STATE, _T], _STATE], state: Any, xs: FSharpList[Any]) -> Any:
    acc: _STATE = state
    xs_1: FSharpList[_T] = xs
    while not FSharpList__get_IsEmpty(xs_1):
        acc = folder(acc, head(xs_1))
        xs_1 = FSharpList__get_Tail(xs_1)
    return acc


def reverse(xs: FSharpList[Any]) -> FSharpList[Any]:
    def _arrow68(acc: FSharpList[_T], x: _T, xs: Any = xs) -> FSharpList[_T]:
        return FSharpList_Cons_305B8EAC(x, acc)

    return fold(_arrow68, FSharpList_get_Empty(), xs)


def fold_back(folder: Callable[[_T, _STATE], _STATE], xs: FSharpList[Any], state: Any) -> Any:
    return fold_back_1(folder, to_array(xs), state)


def fold_indexed(folder: Callable[[int, _STATE, _T], _STATE], state: Any, xs: FSharpList[Any]) -> Any:
    def loop(
        i_mut: int, acc_mut: _STATE, xs_1_mut: FSharpList[_T], folder: Any = folder, state: Any = state, xs: Any = xs
    ) -> _STATE:
        while True:
            (i, acc, xs_1) = (i_mut, acc_mut, xs_1_mut)
            if FSharpList__get_IsEmpty(xs_1):
                return acc

            else:
                i_mut = i + 1
                acc_mut = folder(i, acc, FSharpList__get_Head(xs_1))
                xs_1_mut = FSharpList__get_Tail(xs_1)
                continue

            break

    return loop(0, state, xs)


def fold2(folder: Callable[[_STATE, _T1, _T2], _STATE], state: Any, xs: FSharpList[Any], ys: FSharpList[Any]) -> Any:
    acc: _STATE = state
    xs_1: FSharpList[_T1] = xs
    ys_1: FSharpList[_T2] = ys
    while (not FSharpList__get_IsEmpty(ys_1)) if (not FSharpList__get_IsEmpty(xs_1)) else False:
        acc = folder(acc, FSharpList__get_Head(xs_1), FSharpList__get_Head(ys_1))
        xs_1 = FSharpList__get_Tail(xs_1)
        ys_1 = FSharpList__get_Tail(ys_1)
    return acc


def fold_back2(
    folder: Callable[[_T1, _T2, _STATE], _STATE], xs: FSharpList[Any], ys: FSharpList[Any], state: Any
) -> Any:
    return fold_back2_1(folder, to_array(xs), to_array(ys), state)


def unfold(gen: Callable[[_STATE], tuple[_T, _STATE] | None], state: Any) -> FSharpList[Any]:
    def loop(acc_mut: _STATE, node_mut: FSharpList[_T], gen: Any = gen, state: Any = state) -> FSharpList[_T]:
        while True:
            (acc, node) = (acc_mut, node_mut)
            match_value: tuple[_T, _STATE] | None = gen(acc)
            if match_value is not None:
                acc_mut = match_value[1]

                def _arrow69(__unit: None = None, acc: Any = acc, node: Any = node) -> FSharpList[_T]:
                    t: FSharpList[_T] = FSharpList(match_value[0], None)
                    node.tail = t
                    return t

                node_mut = _arrow69()
                continue

            else:
                return node

            break

    root: FSharpList[_T] = FSharpList_get_Empty()
    node_1: FSharpList[_T] = loop(state, root)
    t_2: FSharpList[_T] = FSharpList_get_Empty()
    node_1.tail = t_2
    return FSharpList__get_Tail(root)


def iterate(action: Callable[[__A], None], xs: FSharpList[Any]) -> None:
    def _arrow70(unit_var: None, x: __A, action: Any = action, xs: Any = xs) -> None:
        action(x)

    fold(_arrow70, None, xs)


def iterate2(action: Callable[[__A, __B], None], xs: FSharpList[Any], ys: FSharpList[Any]) -> None:
    def _arrow71(unit_var: None, x: __A, y: __B, action: Any = action, xs: Any = xs, ys: Any = ys) -> None:
        action(x, y)

    fold2(_arrow71, None, xs, ys)


def iterate_indexed(action: Callable[[int, __A], None], xs: FSharpList[Any]) -> None:
    def _arrow72(i: int, x: __A, action: Any = action, xs: Any = xs) -> int:
        action(i, x)
        return i + 1

    ignore(fold(_arrow72, 0, xs))


def iterate_indexed2(action: Callable[[int, __A, __B], None], xs: FSharpList[Any], ys: FSharpList[Any]) -> None:
    def _arrow73(i: int, x: __A, y: __B, action: Any = action, xs: Any = xs, ys: Any = ys) -> int:
        action(i, x, y)
        return i + 1

    ignore(fold2(_arrow73, 0, xs, ys))


def to_seq(xs: FSharpList[Any]) -> IEnumerable_1[Any]:
    return xs


def of_array_with_tail(xs: Array[Any], tail_1: FSharpList[Any]) -> FSharpList[Any]:
    res: FSharpList[_T] = tail_1
    for i in range(len(xs) - 1, 0 - 1, -1):
        res = FSharpList_Cons_305B8EAC(xs[i], res)
    return res


def of_array(xs: Array[Any]) -> FSharpList[Any]:
    return of_array_with_tail(xs, FSharpList_get_Empty())


def of_seq(xs: IEnumerable_1[Any]) -> FSharpList[Any]:
    if is_array_like(xs):
        return of_array(xs)

    elif isinstance(xs, FSharpList):
        return xs

    else:
        root: FSharpList[_T] = FSharpList_get_Empty()
        node: FSharpList[_T] = root
        with get_enumerator(xs) as enumerator:
            while enumerator.System_Collections_IEnumerator_MoveNext():
                x: _T = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()

                def _arrow74(__unit: None = None, xs: Any = xs) -> FSharpList[_T]:
                    xs_3: FSharpList[_T] = node
                    t: FSharpList[_T] = FSharpList(x, None)
                    xs_3.tail = t
                    return t

                node = _arrow74()
        xs_5: FSharpList[_T] = node
        t_2: FSharpList[_T] = FSharpList_get_Empty()
        xs_5.tail = t_2
        return FSharpList__get_Tail(root)


def concat(lists: IEnumerable_1[FSharpList[Any]]) -> FSharpList[Any]:
    root: FSharpList[_T] = FSharpList_get_Empty()
    node: FSharpList[_T] = root

    def action(xs: FSharpList[_T], lists: Any = lists) -> None:
        nonlocal node

        def _arrow75(acc: FSharpList[_T], x: _T, xs: Any = xs) -> FSharpList[_T]:
            t: FSharpList[_T] = FSharpList(x, None)
            acc.tail = t
            return t

        node = fold(_arrow75, node, xs)

    if is_array_like(lists):
        iterate_1(action, lists)

    elif isinstance(lists, FSharpList):
        iterate(action, lists)

    else:
        with get_enumerator(lists) as enumerator:
            while enumerator.System_Collections_IEnumerator_MoveNext():
                action(enumerator.System_Collections_Generic_IEnumerator_1_get_Current())

    xs_6: FSharpList[_T] = node
    t_2: FSharpList[_T] = FSharpList_get_Empty()
    xs_6.tail = t_2
    return FSharpList__get_Tail(root)


def scan(folder: Callable[[_STATE, _T], _STATE], state: Any, xs: FSharpList[Any]) -> FSharpList[Any]:
    root: FSharpList[_STATE] = FSharpList_get_Empty()
    node: FSharpList[_STATE]
    t: FSharpList[_STATE] = FSharpList(state, None)
    root.tail = t
    node = t
    acc: _STATE = state
    xs_3: FSharpList[_T] = xs
    while not FSharpList__get_IsEmpty(xs_3):
        acc = folder(acc, FSharpList__get_Head(xs_3))

        def _arrow76(__unit: None = None, folder: Any = folder, state: Any = state, xs: Any = xs) -> FSharpList[_STATE]:
            xs_4: FSharpList[_STATE] = node
            t_2: FSharpList[_STATE] = FSharpList(acc, None)
            xs_4.tail = t_2
            return t_2

        node = _arrow76()
        xs_3 = FSharpList__get_Tail(xs_3)
    xs_6: FSharpList[_STATE] = node
    t_4: FSharpList[_STATE] = FSharpList_get_Empty()
    xs_6.tail = t_4
    return FSharpList__get_Tail(root)


def scan_back(folder: Callable[[_T, _STATE], _STATE], xs: FSharpList[Any], state: Any) -> FSharpList[Any]:
    return of_array(scan_back_1(folder, to_array(xs), state, None))


def append(xs: FSharpList[Any], ys: FSharpList[Any]) -> FSharpList[Any]:
    def _arrow77(acc: FSharpList[_T], x: _T, xs: Any = xs, ys: Any = ys) -> FSharpList[_T]:
        return FSharpList_Cons_305B8EAC(x, acc)

    return fold(_arrow77, ys, reverse(xs))


def collect(mapping: Callable[[_T], FSharpList[_U]], xs: FSharpList[Any]) -> FSharpList[Any]:
    root: FSharpList[_U] = FSharpList_get_Empty()
    node: FSharpList[_U] = root
    ys: FSharpList[_T] = xs
    while not FSharpList__get_IsEmpty(ys):
        zs: FSharpList[_U] = mapping(FSharpList__get_Head(ys))
        while not FSharpList__get_IsEmpty(zs):

            def _arrow78(__unit: None = None, mapping: Any = mapping, xs: Any = xs) -> FSharpList[_U]:
                xs_1: FSharpList[_U] = node
                t: FSharpList[_U] = FSharpList(FSharpList__get_Head(zs), None)
                xs_1.tail = t
                return t

            node = _arrow78()
            zs = FSharpList__get_Tail(zs)
        ys = FSharpList__get_Tail(ys)
    xs_3: FSharpList[_U] = node
    t_2: FSharpList[_U] = FSharpList_get_Empty()
    xs_3.tail = t_2
    return FSharpList__get_Tail(root)


def map_indexed(mapping: Callable[[int, _T], _U], xs: FSharpList[Any]) -> FSharpList[Any]:
    root: FSharpList[_U] = FSharpList_get_Empty()

    def folder(i: int, acc: FSharpList[_U], x: _T, mapping: Any = mapping, xs: Any = xs) -> FSharpList[_U]:
        t: FSharpList[_U] = FSharpList(mapping(i, x), None)
        acc.tail = t
        return t

    node: FSharpList[_U] = fold_indexed(folder, root, xs)
    t_2: FSharpList[_U] = FSharpList_get_Empty()
    node.tail = t_2
    return FSharpList__get_Tail(root)


def map(mapping: Callable[[_T], _U], xs: FSharpList[Any]) -> FSharpList[Any]:
    root: FSharpList[_U] = FSharpList_get_Empty()

    def folder(acc: FSharpList[_U], x: _T, mapping: Any = mapping, xs: Any = xs) -> FSharpList[_U]:
        t: FSharpList[_U] = FSharpList(mapping(x), None)
        acc.tail = t
        return t

    node: FSharpList[_U] = fold(folder, root, xs)
    t_2: FSharpList[_U] = FSharpList_get_Empty()
    node.tail = t_2
    return FSharpList__get_Tail(root)


def indexed(xs: FSharpList[Any]) -> FSharpList[tuple[int, __A]]:
    def _arrow80(i: int, x: __A, xs: Any = xs) -> tuple[int, __A]:
        return (i, x)

    return map_indexed(_arrow80, xs)


def map2(mapping: Callable[[_T1, _T2], _U], xs: FSharpList[Any], ys: FSharpList[Any]) -> FSharpList[Any]:
    root: FSharpList[_U] = FSharpList_get_Empty()

    def folder(
        acc: FSharpList[_U], x: _T1, y: _T2, mapping: Any = mapping, xs: Any = xs, ys: Any = ys
    ) -> FSharpList[_U]:
        t: FSharpList[_U] = FSharpList(mapping(x, y), None)
        acc.tail = t
        return t

    node: FSharpList[_U] = fold2(folder, root, xs, ys)
    t_2: FSharpList[_U] = FSharpList_get_Empty()
    node.tail = t_2
    return FSharpList__get_Tail(root)


def map_indexed2(mapping: Callable[[int, _T1, _T2], _U], xs: FSharpList[Any], ys: FSharpList[Any]) -> FSharpList[Any]:
    def loop(
        i_mut: int,
        acc_mut: FSharpList[_U],
        xs_1_mut: FSharpList[_T1],
        ys_1_mut: FSharpList[_T2],
        mapping: Any = mapping,
        xs: Any = xs,
        ys: Any = ys,
    ) -> FSharpList[_U]:
        while True:
            (i, acc, xs_1, ys_1) = (i_mut, acc_mut, xs_1_mut, ys_1_mut)
            if True if FSharpList__get_IsEmpty(xs_1) else FSharpList__get_IsEmpty(ys_1):
                return acc

            else:
                i_mut = i + 1

                def _arrow91(
                    __unit: None = None, i: Any = i, acc: Any = acc, xs_1: Any = xs_1, ys_1: Any = ys_1
                ) -> FSharpList[_U]:
                    t: FSharpList[_U] = FSharpList(
                        mapping(i, FSharpList__get_Head(xs_1), FSharpList__get_Head(ys_1)), None
                    )
                    acc.tail = t
                    return t

                acc_mut = _arrow91()
                xs_1_mut = FSharpList__get_Tail(xs_1)
                ys_1_mut = FSharpList__get_Tail(ys_1)
                continue

            break

    root: FSharpList[_U] = FSharpList_get_Empty()
    node_1: FSharpList[_U] = loop(0, root, xs, ys)
    t_2: FSharpList[_U] = FSharpList_get_Empty()
    node_1.tail = t_2
    return FSharpList__get_Tail(root)


def map3(
    mapping: Callable[[_T1, _T2, _T3], _U], xs: FSharpList[Any], ys: FSharpList[Any], zs: FSharpList[Any]
) -> FSharpList[Any]:
    def loop(
        acc_mut: FSharpList[_U],
        xs_1_mut: FSharpList[_T1],
        ys_1_mut: FSharpList[_T2],
        zs_1_mut: FSharpList[_T3],
        mapping: Any = mapping,
        xs: Any = xs,
        ys: Any = ys,
        zs: Any = zs,
    ) -> FSharpList[_U]:
        while True:
            (acc, xs_1, ys_1, zs_1) = (acc_mut, xs_1_mut, ys_1_mut, zs_1_mut)
            if (
                True
                if (True if FSharpList__get_IsEmpty(xs_1) else FSharpList__get_IsEmpty(ys_1))
                else FSharpList__get_IsEmpty(zs_1)
            ):
                return acc

            else:

                def _arrow92(
                    __unit: None = None, acc: Any = acc, xs_1: Any = xs_1, ys_1: Any = ys_1, zs_1: Any = zs_1
                ) -> FSharpList[_U]:
                    t: FSharpList[_U] = FSharpList(
                        mapping(FSharpList__get_Head(xs_1), FSharpList__get_Head(ys_1), FSharpList__get_Head(zs_1)),
                        None,
                    )
                    acc.tail = t
                    return t

                acc_mut = _arrow92()
                xs_1_mut = FSharpList__get_Tail(xs_1)
                ys_1_mut = FSharpList__get_Tail(ys_1)
                zs_1_mut = FSharpList__get_Tail(zs_1)
                continue

            break

    root: FSharpList[_U] = FSharpList_get_Empty()
    node_1: FSharpList[_U] = loop(root, xs, ys, zs)
    t_2: FSharpList[_U] = FSharpList_get_Empty()
    node_1.tail = t_2
    return FSharpList__get_Tail(root)


def map_fold(
    mapping: Callable[[_STATE, _T], tuple[_RESULT, _STATE]], state: Any, xs: FSharpList[Any]
) -> tuple[FSharpList[_RESULT], _STATE]:
    root: FSharpList[_RESULT] = FSharpList_get_Empty()

    def folder(
        tupled_arg: tuple[FSharpList[_RESULT], _STATE], x: _T, mapping: Any = mapping, state: Any = state, xs: Any = xs
    ) -> tuple[FSharpList[_RESULT], _STATE]:
        pattern_input: tuple[_RESULT, _STATE] = mapping(tupled_arg[1], x)

        def _arrow93(__unit: None = None, tupled_arg: Any = tupled_arg, x: Any = x) -> FSharpList[_RESULT]:
            t: FSharpList[_RESULT] = FSharpList(pattern_input[0], None)
            tupled_arg[0].tail = t
            return t

        return (_arrow93(), pattern_input[1])

    pattern_input_1: tuple[FSharpList[_RESULT], _STATE] = fold(folder, (root, state), xs)
    t_2: FSharpList[_RESULT] = FSharpList_get_Empty()
    pattern_input_1[0].tail = t_2
    return (FSharpList__get_Tail(root), pattern_input_1[1])


def map_fold_back(
    mapping: Callable[[_T, _STATE], tuple[_RESULT, _STATE]], xs: FSharpList[Any], state: Any
) -> tuple[FSharpList[_RESULT], _STATE]:
    def _arrow94(
        acc: _STATE, x: _T, mapping: Any = mapping, xs: Any = xs, state: Any = state
    ) -> tuple[_RESULT, _STATE]:
        return mapping(x, acc)

    return map_fold(_arrow94, state, reverse(xs))


def try_pick(f: Callable[[_T], __A | None], xs: FSharpList[Any]) -> Any | None:
    def loop(xs_1_mut: FSharpList[_T], f: Any = f, xs: Any = xs) -> __A | None:
        while True:
            (xs_1,) = (xs_1_mut,)
            if FSharpList__get_IsEmpty(xs_1):
                return None

            else:
                match_value: __A | None = f(FSharpList__get_Head(xs_1))
                if match_value is None:
                    xs_1_mut = FSharpList__get_Tail(xs_1)
                    continue

                else:
                    return match_value

            break

    return loop(xs)


def pick(f: Callable[[__A], __B | None], xs: FSharpList[Any]) -> Any:
    match_value: __B | None = try_pick(f, xs)
    if match_value is None:
        return index_not_found()

    else:
        return value_1(match_value)


def try_find(f: Callable[[__A], bool], xs: FSharpList[Any]) -> Any | None:
    def _arrow95(x: __A | None = None, f: Any = f, xs: Any = xs) -> __A | None:
        return some(x) if f(x) else None

    return try_pick(_arrow95, xs)


def find(f: Callable[[__A], bool], xs: FSharpList[Any]) -> Any:
    match_value: __A | None = try_find(f, xs)
    if match_value is None:
        return index_not_found()

    else:
        return value_1(match_value)


def try_find_back(f: Callable[[__A], bool], xs: FSharpList[Any]) -> Any | None:
    return try_find_back_1(f, to_array(xs))


def find_back(f: Callable[[__A], bool], xs: FSharpList[Any]) -> Any:
    match_value: __A | None = try_find_back(f, xs)
    if match_value is None:
        return index_not_found()

    else:
        return value_1(match_value)


def try_find_index(f: Callable[[_T], bool], xs: FSharpList[Any]) -> int | None:
    def loop(i_mut: int, xs_1_mut: FSharpList[_T], f: Any = f, xs: Any = xs) -> int | None:
        while True:
            (i, xs_1) = (i_mut, xs_1_mut)
            if FSharpList__get_IsEmpty(xs_1):
                return None

            elif f(FSharpList__get_Head(xs_1)):
                return i

            else:
                i_mut = i + 1
                xs_1_mut = FSharpList__get_Tail(xs_1)
                continue

            break

    return loop(0, xs)


def find_index(f: Callable[[__A], bool], xs: FSharpList[Any]) -> int:
    match_value: int | None = try_find_index(f, xs)
    if match_value is None:
        index_not_found()
        return -1

    else:
        return match_value


def try_find_index_back(f: Callable[[__A], bool], xs: FSharpList[Any]) -> int | None:
    return try_find_index_back_1(f, to_array(xs))


def find_index_back(f: Callable[[__A], bool], xs: FSharpList[Any]) -> int:
    match_value: int | None = try_find_index_back(f, xs)
    if match_value is None:
        index_not_found()
        return -1

    else:
        return match_value


def try_item(n: int, xs: FSharpList[Any]) -> Any | None:
    def loop(i_mut: int, xs_1_mut: FSharpList[_T], n: Any = n, xs: Any = xs) -> _T | None:
        while True:
            (i, xs_1) = (i_mut, xs_1_mut)
            if FSharpList__get_IsEmpty(xs_1):
                return None

            elif i == n:
                return some(FSharpList__get_Head(xs_1))

            else:
                i_mut = i + 1
                xs_1_mut = FSharpList__get_Tail(xs_1)
                continue

            break

    return loop(0, xs)


def item(n: int, xs: FSharpList[Any]) -> Any:
    return FSharpList__get_Item_Z524259A4(xs, n)


def filter(f: Callable[[_T], bool], xs: FSharpList[Any]) -> FSharpList[Any]:
    root: FSharpList[_T] = FSharpList_get_Empty()

    def folder(acc: FSharpList[_T], x: _T, f: Any = f, xs: Any = xs) -> FSharpList[_T]:
        if f(x):
            t: FSharpList[_T] = FSharpList(x, None)
            acc.tail = t
            return t

        else:
            return acc

    node: FSharpList[_T] = fold(folder, root, xs)
    t_2: FSharpList[_T] = FSharpList_get_Empty()
    node.tail = t_2
    return FSharpList__get_Tail(root)


def partition(f: Callable[[_T], bool], xs: FSharpList[Any]) -> tuple[FSharpList[_T], FSharpList[_T]]:
    match_value: FSharpList[_T] = FSharpList_get_Empty()
    root2: FSharpList[_T] = FSharpList_get_Empty()
    root1: FSharpList[_T] = match_value

    def folder(
        tupled_arg: tuple[FSharpList[_T], FSharpList[_T]], x: _T, f: Any = f, xs: Any = xs
    ) -> tuple[FSharpList[_T], FSharpList[_T]]:
        lacc: FSharpList[_T] = tupled_arg[0]
        racc: FSharpList[_T] = tupled_arg[1]
        if f(x):

            def _arrow96(__unit: None = None, tupled_arg: Any = tupled_arg, x: Any = x) -> FSharpList[_T]:
                t: FSharpList[_T] = FSharpList(x, None)
                lacc.tail = t
                return t

            return (_arrow96(), racc)

        else:

            def _arrow97(__unit: None = None, tupled_arg: Any = tupled_arg, x: Any = x) -> FSharpList[_T]:
                t_2: FSharpList[_T] = FSharpList(x, None)
                racc.tail = t_2
                return t_2

            return (lacc, _arrow97())

    pattern_input_1: tuple[FSharpList[_T], FSharpList[_T]] = fold(folder, (root1, root2), xs)
    t_4: FSharpList[_T] = FSharpList_get_Empty()
    pattern_input_1[0].tail = t_4
    t_5: FSharpList[_T] = FSharpList_get_Empty()
    pattern_input_1[1].tail = t_5
    return (FSharpList__get_Tail(root1), FSharpList__get_Tail(root2))


def choose(f: Callable[[_T], _U | None], xs: FSharpList[Any]) -> FSharpList[Any]:
    root: FSharpList[_U] = FSharpList_get_Empty()

    def folder(acc: FSharpList[_U], x: _T, f: Any = f, xs: Any = xs) -> FSharpList[_U]:
        match_value: _U | None = f(x)
        if match_value is None:
            return acc

        else:
            t: FSharpList[_U] = FSharpList(value_1(match_value), None)
            acc.tail = t
            return t

    node: FSharpList[_U] = fold(folder, root, xs)
    t_2: FSharpList[_U] = FSharpList_get_Empty()
    node.tail = t_2
    return FSharpList__get_Tail(root)


def contains(value: Any, xs: FSharpList[Any], eq: IEqualityComparer_1[Any]) -> bool:
    def _arrow98(v: _T | None = None, value: Any = value, xs: Any = xs, eq: Any = eq) -> bool:
        return eq.Equals(value, v)

    return try_find_index(_arrow98, xs) is not None


def initialize(n: int, f: Callable[[int], _T]) -> FSharpList[Any]:
    root: FSharpList[_T] = FSharpList_get_Empty()
    node: FSharpList[_T] = root
    for i in range(0, (n - 1) + 1, 1):

        def _arrow99(__unit: None = None, n: Any = n, f: Any = f) -> FSharpList[_T]:
            xs: FSharpList[_T] = node
            t: FSharpList[_T] = FSharpList(f(i), None)
            xs.tail = t
            return t

        node = _arrow99()
    xs_2: FSharpList[_T] = node
    t_2: FSharpList[_T] = FSharpList_get_Empty()
    xs_2.tail = t_2
    return FSharpList__get_Tail(root)


def replicate(n: int, x: Any) -> FSharpList[Any]:
    def _arrow103(_arg: int, n: Any = n, x: Any = x) -> __A:
        return x

    return initialize(n, _arrow103)


def reduce(f: Callable[[_T, _T], _T], xs: FSharpList[Any]) -> Any:
    if FSharpList__get_IsEmpty(xs):
        raise Exception(SR_inputWasEmpty)

    else:
        return fold(f, head(xs), tail(xs))


def reduce_back(f: Callable[[_T, _T], _T], xs: FSharpList[Any]) -> Any:
    if FSharpList__get_IsEmpty(xs):
        raise Exception(SR_inputWasEmpty)

    else:
        return fold_back(f, tail(xs), head(xs))


def for_all(f: Callable[[__A], bool], xs: FSharpList[Any]) -> bool:
    def _arrow108(acc: bool, x: __A, f: Any = f, xs: Any = xs) -> bool:
        return f(x) if acc else False

    return fold(_arrow108, True, xs)


def for_all2(f: Callable[[__A, __B], bool], xs: FSharpList[Any], ys: FSharpList[Any]) -> bool:
    def _arrow110(acc: bool, x: __A, y: __B, f: Any = f, xs: Any = xs, ys: Any = ys) -> bool:
        return f(x, y) if acc else False

    return fold2(_arrow110, True, xs, ys)


def exists(f: Callable[[__A], bool], xs: FSharpList[Any]) -> bool:
    return try_find_index(f, xs) is not None


def exists2(f_mut: Callable[[_T1, _T2], bool], xs_mut: FSharpList[Any], ys_mut: FSharpList[Any]) -> bool:
    while True:
        (f, xs, ys) = (f_mut, xs_mut, ys_mut)
        match_value: bool = FSharpList__get_IsEmpty(xs)
        match_value_1: bool = FSharpList__get_IsEmpty(ys)
        (pattern_matching_result,) = (None,)
        if match_value:
            if match_value_1:
                pattern_matching_result = 0

            else:
                pattern_matching_result = 2

        elif match_value_1:
            pattern_matching_result = 2

        else:
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return False

        elif pattern_matching_result == 1:
            if f(FSharpList__get_Head(xs), FSharpList__get_Head(ys)):
                return True

            else:
                f_mut = f
                xs_mut = FSharpList__get_Tail(xs)
                ys_mut = FSharpList__get_Tail(ys)
                continue

        elif pattern_matching_result == 2:
            raise Exception((SR_differentLengths + "\\nParameter name: ") + "list2")

        break


def unzip(xs: FSharpList[tuple[__A, __B]]) -> tuple[FSharpList[__A], FSharpList[__B]]:
    def _arrow111(
        tupled_arg: tuple[__A, __B], tupled_arg_1: tuple[FSharpList[__A], FSharpList[__B]], xs: Any = xs
    ) -> tuple[FSharpList[__A], FSharpList[__B]]:
        return (
            FSharpList_Cons_305B8EAC(tupled_arg[0], tupled_arg_1[0]),
            FSharpList_Cons_305B8EAC(tupled_arg[1], tupled_arg_1[1]),
        )

    return fold_back(_arrow111, xs, (FSharpList_get_Empty(), FSharpList_get_Empty()))


def unzip3(xs: FSharpList[tuple[__A, __B, __C]]) -> tuple[FSharpList[__A], FSharpList[__B], FSharpList[__C]]:
    def _arrow112(
        tupled_arg: tuple[__A, __B, __C],
        tupled_arg_1: tuple[FSharpList[__A], FSharpList[__B], FSharpList[__C]],
        xs: Any = xs,
    ) -> tuple[FSharpList[__A], FSharpList[__B], FSharpList[__C]]:
        return (
            FSharpList_Cons_305B8EAC(tupled_arg[0], tupled_arg_1[0]),
            FSharpList_Cons_305B8EAC(tupled_arg[1], tupled_arg_1[1]),
            FSharpList_Cons_305B8EAC(tupled_arg[2], tupled_arg_1[2]),
        )

    return fold_back(_arrow112, xs, (FSharpList_get_Empty(), FSharpList_get_Empty(), FSharpList_get_Empty()))


def zip(xs: FSharpList[Any], ys: FSharpList[Any]) -> FSharpList[tuple[__A, __B]]:
    def _arrow113(x: __A, y: __B, xs: Any = xs, ys: Any = ys) -> tuple[__A, __B]:
        return (x, y)

    return map2(_arrow113, xs, ys)


def zip3(xs: FSharpList[Any], ys: FSharpList[Any], zs: FSharpList[Any]) -> FSharpList[tuple[__A, __B, __C]]:
    def _arrow114(x: __A, y: __B, z: __C, xs: Any = xs, ys: Any = ys, zs: Any = zs) -> tuple[__A, __B, __C]:
        return (x, y, z)

    return map3(_arrow114, xs, ys, zs)


def sort_with(comparer: Callable[[_T, _T], int], xs: FSharpList[Any]) -> FSharpList[Any]:
    arr: Array[_T] = to_array(xs)
    sort_in_place_with(comparer, arr)
    return of_array(arr)


def sort(xs: FSharpList[Any], comparer: IComparer_1[Any]) -> FSharpList[Any]:
    def _arrow115(x: _T, y: _T, xs: Any = xs, comparer: Any = comparer) -> int:
        return comparer.Compare(x, y)

    return sort_with(_arrow115, xs)


def sort_by(projection: Callable[[_T], _U], xs: FSharpList[Any], comparer: IComparer_1[Any]) -> FSharpList[Any]:
    def _arrow116(x: _T, y: _T, projection: Any = projection, xs: Any = xs, comparer: Any = comparer) -> int:
        return comparer.Compare(projection(x), projection(y))

    return sort_with(_arrow116, xs)


def sort_descending(xs: FSharpList[Any], comparer: IComparer_1[Any]) -> FSharpList[Any]:
    def _arrow117(x: _T, y: _T, xs: Any = xs, comparer: Any = comparer) -> int:
        return comparer.Compare(x, y) * -1

    return sort_with(_arrow117, xs)


def sort_by_descending(
    projection: Callable[[_T], _U], xs: FSharpList[Any], comparer: IComparer_1[Any]
) -> FSharpList[Any]:
    def _arrow118(x: _T, y: _T, projection: Any = projection, xs: Any = xs, comparer: Any = comparer) -> int:
        return comparer.Compare(projection(x), projection(y)) * -1

    return sort_with(_arrow118, xs)


def sum(xs: FSharpList[Any], adder: IGenericAdder_1[Any]) -> Any:
    def _arrow119(acc: _T, x: _T, xs: Any = xs, adder: Any = adder) -> _T:
        return adder.Add(acc, x)

    return fold(_arrow119, adder.GetZero(), xs)


def sum_by(f: Callable[[_T], _U], xs: FSharpList[Any], adder: IGenericAdder_1[Any]) -> Any:
    def _arrow120(acc: _U, x: _T, f: Any = f, xs: Any = xs, adder: Any = adder) -> _U:
        return adder.Add(acc, f(x))

    return fold(_arrow120, adder.GetZero(), xs)


def max_by(projection: Callable[[_T], _U], xs: FSharpList[Any], comparer: IComparer_1[Any]) -> Any:
    def _arrow121(x: _T, y: _T, projection: Any = projection, xs: Any = xs, comparer: Any = comparer) -> _T:
        return y if (comparer.Compare(projection(y), projection(x)) > 0) else x

    return reduce(_arrow121, xs)


def max(xs: FSharpList[Any], comparer: IComparer_1[Any]) -> Any:
    def _arrow122(x: _T, y: _T, xs: Any = xs, comparer: Any = comparer) -> _T:
        return y if (comparer.Compare(y, x) > 0) else x

    return reduce(_arrow122, xs)


def min_by(projection: Callable[[_T], _U], xs: FSharpList[Any], comparer: IComparer_1[Any]) -> Any:
    def _arrow123(x: _T, y: _T, projection: Any = projection, xs: Any = xs, comparer: Any = comparer) -> _T:
        return x if (comparer.Compare(projection(y), projection(x)) > 0) else y

    return reduce(_arrow123, xs)


def min(xs: FSharpList[Any], comparer: IComparer_1[Any]) -> Any:
    def _arrow124(x: _T, y: _T, xs: Any = xs, comparer: Any = comparer) -> _T:
        return x if (comparer.Compare(y, x) > 0) else y

    return reduce(_arrow124, xs)


def average(xs: FSharpList[Any], averager: IGenericAverager_1[Any]) -> Any:
    count: int = 0

    def folder(acc: _T, x: _T, xs: Any = xs, averager: Any = averager) -> _T:
        nonlocal count
        count = (count + 1) or 0
        return averager.Add(acc, x)

    total: _T = fold(folder, averager.GetZero(), xs)
    return averager.DivideByInt(total, count)


def average_by(f: Callable[[_T], _U], xs: FSharpList[Any], averager: IGenericAverager_1[Any]) -> Any:
    count: int = 0

    def _arrow125(acc: _U, x: _T, f: Any = f, xs: Any = xs, averager: Any = averager) -> _U:
        nonlocal count
        count = (count + 1) or 0
        return averager.Add(acc, f(x))

    total: _U = fold(_arrow125, averager.GetZero(), xs)
    return averager.DivideByInt(total, count)


def permute(f: Callable[[int], int], xs: FSharpList[Any]) -> FSharpList[Any]:
    return of_array(permute_1(f, to_array(xs)))


def chunk_by_size(chunk_size: int, xs: FSharpList[Any]) -> FSharpList[FSharpList[Any]]:
    def mapping(xs_1: Array[_T], chunk_size: Any = chunk_size, xs: Any = xs) -> FSharpList[_T]:
        return of_array(xs_1)

    return of_array(map_1(mapping, chunk_by_size_1(chunk_size, to_array(xs)), None))


def all_pairs(xs: FSharpList[Any], ys: FSharpList[Any]) -> FSharpList[tuple[_T1, _T2]]:
    root: FSharpList[tuple[_T1, _T2]] = FSharpList_get_Empty()
    node: FSharpList[tuple[_T1, _T2]] = root

    def _arrow128(x: _T1 | None = None, xs: Any = xs, ys: Any = ys) -> None:
        def _arrow127(y: _T2 | None = None) -> None:
            nonlocal node

            def _arrow126(__unit: None = None) -> FSharpList[tuple[_T1, _T2]]:
                xs_1: FSharpList[tuple[_T1, _T2]] = node
                t: FSharpList[tuple[_T1, _T2]] = FSharpList((x, y), None)
                xs_1.tail = t
                return t

            node = _arrow126()

        iterate(_arrow127, ys)

    iterate(_arrow128, xs)
    xs_3: FSharpList[tuple[_T1, _T2]] = node
    t_2: FSharpList[tuple[_T1, _T2]] = FSharpList_get_Empty()
    xs_3.tail = t_2
    return FSharpList__get_Tail(root)


def skip(count_mut: int, xs_mut: FSharpList[Any]) -> FSharpList[Any]:
    while True:
        (count, xs) = (count_mut, xs_mut)
        if count <= 0:
            return xs

        elif FSharpList__get_IsEmpty(xs):
            raise Exception((SR_notEnoughElements + "\\nParameter name: ") + "list")

        else:
            count_mut = count - 1
            xs_mut = FSharpList__get_Tail(xs)
            continue

        break


def skip_while(predicate_mut: Callable[[_T], bool], xs_mut: FSharpList[Any]) -> FSharpList[Any]:
    while True:
        (predicate, xs) = (predicate_mut, xs_mut)
        if FSharpList__get_IsEmpty(xs):
            return xs

        elif not predicate(FSharpList__get_Head(xs)):
            return xs

        else:
            predicate_mut = predicate
            xs_mut = FSharpList__get_Tail(xs)
            continue

        break


def take(count: int, xs: FSharpList[Any]) -> FSharpList[Any]:
    if count < 0:
        raise Exception((SR_inputMustBeNonNegative + "\\nParameter name: ") + "count")

    def loop(
        i_mut: int, acc_mut: FSharpList[_T], xs_1_mut: FSharpList[_T], count: Any = count, xs: Any = xs
    ) -> FSharpList[_T]:
        while True:
            (i, acc, xs_1) = (i_mut, acc_mut, xs_1_mut)
            if i <= 0:
                return acc

            elif FSharpList__get_IsEmpty(xs_1):
                raise Exception((SR_notEnoughElements + "\\nParameter name: ") + "list")

            else:
                i_mut = i - 1

                def _arrow129(__unit: None = None, i: Any = i, acc: Any = acc, xs_1: Any = xs_1) -> FSharpList[_T]:
                    t: FSharpList[_T] = FSharpList(FSharpList__get_Head(xs_1), None)
                    acc.tail = t
                    return t

                acc_mut = _arrow129()
                xs_1_mut = FSharpList__get_Tail(xs_1)
                continue

            break

    root: FSharpList[_T] = FSharpList_get_Empty()
    node: FSharpList[_T] = loop(count, root, xs)
    t_2: FSharpList[_T] = FSharpList_get_Empty()
    node.tail = t_2
    return FSharpList__get_Tail(root)


def take_while(predicate: Callable[[_T], bool], xs: FSharpList[Any]) -> FSharpList[Any]:
    def loop(
        acc_mut: FSharpList[_T], xs_1_mut: FSharpList[_T], predicate: Any = predicate, xs: Any = xs
    ) -> FSharpList[_T]:
        while True:
            (acc, xs_1) = (acc_mut, xs_1_mut)
            if FSharpList__get_IsEmpty(xs_1):
                return acc

            elif not predicate(FSharpList__get_Head(xs_1)):
                return acc

            else:

                def _arrow130(__unit: None = None, acc: Any = acc, xs_1: Any = xs_1) -> FSharpList[_T]:
                    t: FSharpList[_T] = FSharpList(FSharpList__get_Head(xs_1), None)
                    acc.tail = t
                    return t

                acc_mut = _arrow130()
                xs_1_mut = FSharpList__get_Tail(xs_1)
                continue

            break

    root: FSharpList[_T] = FSharpList_get_Empty()
    node: FSharpList[_T] = loop(root, xs)
    t_2: FSharpList[_T] = FSharpList_get_Empty()
    node.tail = t_2
    return FSharpList__get_Tail(root)


def truncate(count: int, xs: FSharpList[Any]) -> FSharpList[Any]:
    def loop(
        i_mut: int, acc_mut: FSharpList[_T], xs_1_mut: FSharpList[_T], count: Any = count, xs: Any = xs
    ) -> FSharpList[_T]:
        while True:
            (i, acc, xs_1) = (i_mut, acc_mut, xs_1_mut)
            if i <= 0:
                return acc

            elif FSharpList__get_IsEmpty(xs_1):
                return acc

            else:
                i_mut = i - 1

                def _arrow131(__unit: None = None, i: Any = i, acc: Any = acc, xs_1: Any = xs_1) -> FSharpList[_T]:
                    t: FSharpList[_T] = FSharpList(FSharpList__get_Head(xs_1), None)
                    acc.tail = t
                    return t

                acc_mut = _arrow131()
                xs_1_mut = FSharpList__get_Tail(xs_1)
                continue

            break

    root: FSharpList[_T] = FSharpList_get_Empty()
    node: FSharpList[_T] = loop(count, root, xs)
    t_2: FSharpList[_T] = FSharpList_get_Empty()
    node.tail = t_2
    return FSharpList__get_Tail(root)


def get_slice(start_index: int | None, end_index: int | None, xs: FSharpList[Any]) -> FSharpList[Any]:
    len_1: int = length(xs) or 0
    start_index_1: int
    index: int = default_arg(start_index, 0) or 0
    start_index_1 = 0 if (index < 0) else index
    end_index_1: int
    index_1: int = default_arg(end_index, len_1 - 1) or 0
    end_index_1 = (len_1 - 1) if (index_1 >= len_1) else index_1
    if end_index_1 < start_index_1:
        return FSharpList_get_Empty()

    else:
        return take((end_index_1 - start_index_1) + 1, skip(start_index_1, xs))


def split_at(index: int, xs: FSharpList[Any]) -> tuple[FSharpList[_T], FSharpList[_T]]:
    if index < 0:
        raise Exception((SR_inputMustBeNonNegative + "\\nParameter name: ") + "index")

    if index > FSharpList__get_Length(xs):
        raise Exception((SR_notEnoughElements + "\\nParameter name: ") + "index")

    return (take(index, xs), skip(index, xs))


def exactly_one(xs: FSharpList[Any]) -> Any:
    if FSharpList__get_IsEmpty(xs):
        raise Exception((SR_inputSequenceEmpty + "\\nParameter name: ") + "list")

    elif FSharpList__get_IsEmpty(FSharpList__get_Tail(xs)):
        return FSharpList__get_Head(xs)

    else:
        raise Exception((SR_inputSequenceTooLong + "\\nParameter name: ") + "list")


def try_exactly_one(xs: FSharpList[Any]) -> Any | None:
    if FSharpList__get_IsEmpty(FSharpList__get_Tail(xs)) if (not FSharpList__get_IsEmpty(xs)) else False:
        return some(FSharpList__get_Head(xs))

    else:
        return None


def where(predicate: Callable[[_T], bool], xs: FSharpList[Any]) -> FSharpList[Any]:
    return filter(predicate, xs)


def pairwise(xs: FSharpList[Any]) -> FSharpList[tuple[_T, _T]]:
    return of_array(pairwise_1(to_array(xs)))


def windowed(window_size: int, xs: FSharpList[Any]) -> FSharpList[FSharpList[Any]]:
    def mapping(xs_1: Array[_T], window_size: Any = window_size, xs: Any = xs) -> FSharpList[_T]:
        return of_array(xs_1)

    return of_array(map_1(mapping, windowed_1(window_size, to_array(xs)), None))


def split_into(chunks: int, xs: FSharpList[Any]) -> FSharpList[FSharpList[Any]]:
    def mapping(xs_1: Array[_T], chunks: Any = chunks, xs: Any = xs) -> FSharpList[_T]:
        return of_array(xs_1)

    return of_array(map_1(mapping, split_into_1(chunks, to_array(xs)), None))


def transpose(lists: IEnumerable_1[FSharpList[Any]]) -> FSharpList[FSharpList[Any]]:
    def mapping_1(xs_1: Array[_T], lists: Any = lists) -> FSharpList[_T]:
        return of_array(xs_1)

    def mapping(xs: FSharpList[_T], lists: Any = lists) -> Array[_T]:
        return to_array(xs)

    return of_array(map_1(mapping_1, transpose_1(map_1(mapping, list(lists), None), None), None))


def insert_at(index: int, y: Any, xs: FSharpList[Any]) -> FSharpList[Any]:
    i: int = -1
    is_done: bool = False

    def folder(acc: FSharpList[_T], x: _T, index: Any = index, y: Any = y, xs: Any = xs) -> FSharpList[_T]:
        nonlocal i, is_done
        i = (i + 1) or 0
        if i == index:
            is_done = True
            return FSharpList_Cons_305B8EAC(x, FSharpList_Cons_305B8EAC(y, acc))

        else:
            return FSharpList_Cons_305B8EAC(x, acc)

    result: FSharpList[_T] = fold(folder, FSharpList_get_Empty(), xs)

    def _arrow132(__unit: None = None, index: Any = index, y: Any = y, xs: Any = xs) -> FSharpList[_T]:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")

    return reverse(result if is_done else (FSharpList_Cons_305B8EAC(y, result) if ((i + 1) == index) else _arrow132()))


def insert_many_at(index: int, ys: IEnumerable_1[Any], xs: FSharpList[Any]) -> FSharpList[Any]:
    i: int = -1
    is_done: bool = False
    ys_1: FSharpList[_T] = of_seq(ys)

    def folder(acc: FSharpList[_T], x: _T, index: Any = index, ys: Any = ys, xs: Any = xs) -> FSharpList[_T]:
        nonlocal i, is_done
        i = (i + 1) or 0
        if i == index:
            is_done = True
            return FSharpList_Cons_305B8EAC(x, append(ys_1, acc))

        else:
            return FSharpList_Cons_305B8EAC(x, acc)

    result: FSharpList[_T] = fold(folder, FSharpList_get_Empty(), xs)

    def _arrow133(__unit: None = None, index: Any = index, ys: Any = ys, xs: Any = xs) -> FSharpList[_T]:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")

    return reverse(result if is_done else (append(ys_1, result) if ((i + 1) == index) else _arrow133()))


def remove_at(index: int, xs: FSharpList[Any]) -> FSharpList[Any]:
    i: int = -1
    is_done: bool = False

    def f(_arg: _T | None = None, index: Any = index, xs: Any = xs) -> bool:
        nonlocal i, is_done
        i = (i + 1) or 0
        if i == index:
            is_done = True
            return False

        else:
            return True

    ys: FSharpList[_T] = filter(f, xs)
    if not is_done:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")

    return ys


def remove_many_at(index: int, count: int, xs: FSharpList[Any]) -> FSharpList[Any]:
    i: int = -1
    status: int = -1

    def f(_arg: _T | None = None, index: Any = index, count: Any = count, xs: Any = xs) -> bool:
        nonlocal i, status
        i = (i + 1) or 0
        if i == index:
            status = 0
            return False

        elif i > index:
            if i < (index + count):
                return False

            else:
                status = 1
                return True

        else:
            return True

    ys: FSharpList[_T] = filter(f, xs)
    status_1: int = (1 if (((i + 1) == (index + count)) if (status == 0) else False) else status) or 0
    if status_1 < 1:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + ("index" if (status_1 < 0) else "count"))

    return ys


def update_at(index: int, y: Any, xs: FSharpList[Any]) -> FSharpList[Any]:
    is_done: bool = False

    def mapping(i: int, x: _T, index: Any = index, y: Any = y, xs: Any = xs) -> _T:
        nonlocal is_done
        if i == index:
            is_done = True
            return y

        else:
            return x

    ys: FSharpList[_T] = map_indexed(mapping, xs)
    if not is_done:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")

    return ys


__all__ = [
    "FSharpList_reflection",
    "ListEnumerator_1_reflection",
    "FSharpList_get_Empty",
    "FSharpList_Cons_305B8EAC",
    "FSharpList__get_IsEmpty",
    "FSharpList__get_Length",
    "FSharpList__get_Head",
    "FSharpList__get_Tail",
    "FSharpList__get_Item_Z524259A4",
    "index_not_found",
    "empty",
    "cons",
    "singleton",
    "is_empty",
    "length",
    "head",
    "try_head",
    "tail",
    "try_last",
    "last",
    "compare_with",
    "to_array",
    "fold",
    "reverse",
    "fold_back",
    "fold_indexed",
    "fold2",
    "fold_back2",
    "unfold",
    "iterate",
    "iterate2",
    "iterate_indexed",
    "iterate_indexed2",
    "to_seq",
    "of_array_with_tail",
    "of_array",
    "of_seq",
    "concat",
    "scan",
    "scan_back",
    "append",
    "collect",
    "map_indexed",
    "map",
    "indexed",
    "map2",
    "map_indexed2",
    "map3",
    "map_fold",
    "map_fold_back",
    "try_pick",
    "pick",
    "try_find",
    "find",
    "try_find_back",
    "find_back",
    "try_find_index",
    "find_index",
    "try_find_index_back",
    "find_index_back",
    "try_item",
    "item",
    "filter",
    "partition",
    "choose",
    "contains",
    "initialize",
    "replicate",
    "reduce",
    "reduce_back",
    "for_all",
    "for_all2",
    "exists",
    "exists2",
    "unzip",
    "unzip3",
    "zip",
    "zip3",
    "sort_with",
    "sort",
    "sort_by",
    "sort_descending",
    "sort_by_descending",
    "sum",
    "sum_by",
    "max_by",
    "max",
    "min_by",
    "min",
    "average",
    "average_by",
    "permute",
    "chunk_by_size",
    "all_pairs",
    "skip",
    "skip_while",
    "take",
    "take_while",
    "truncate",
    "get_slice",
    "split_at",
    "exactly_one",
    "try_exactly_one",
    "where",
    "pairwise",
    "windowed",
    "split_into",
    "transpose",
    "insert_at",
    "insert_many_at",
    "remove_at",
    "remove_many_at",
    "update_at",
]



from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.seq import (length, min_by, max_by, map, exists)
from ..fable_library.util import (IEnumerable_1, compare_primitives, ignore, get_enumerator, IEnumerator, to_iterator)
from .Cells.fs_cell import FsCell
from .Cells.fs_cells_collection import FsCellsCollection
from .fs_address import FsAddress
from .Ranges.fs_range_address import FsRangeAddress
from .Ranges.fs_range_base import (FsRangeBase, FsRangeBase_reflection)

def _expr233() -> TypeInfo:
    return class_type("FsSpreadsheet.FsColumn", None, FsColumn, FsRangeBase_reflection())


class FsColumn(FsRangeBase):
    def __init__(self, range_address: FsRangeAddress, cells: FsCellsCollection) -> None:
        super().__init__(range_address)
        self.cells_004018: FsCellsCollection = cells

    @staticmethod
    def empty(__unit: None=None) -> FsColumn:
        return FsColumn(FsRangeAddress(FsAddress(0, 0), FsAddress(0, 0)), FsCellsCollection())

    @staticmethod
    def create_at(index: int, cells: FsCellsCollection) -> FsColumn:
        def get_index_by(f: Callable[[Callable[[FsCell], int], IEnumerable_1[FsCell]], FsCell]) -> int:
            if length(cells.GetCellsInColumn(index)) == 0:
                return 1

            else: 
                def _arrow226(c: FsCell, f: Any=f) -> int:
                    return c.Address.RowNumber

                return f(_arrow226)(cells.GetCellsInColumn(index)).Address.RowNumber


        def _arrow229(projection: Callable[[FsCell], int]) -> Callable[[IEnumerable_1[FsCell]], FsCell]:
            def _arrow228(source_1: IEnumerable_1[FsCell]) -> FsCell:
                class ObjectExpr227:
                    @property
                    def Compare(self) -> Callable[[int, int], int]:
                        return compare_primitives

                return min_by(projection, source_1, ObjectExpr227())

            return _arrow228

        min_row_index: int = get_index_by(_arrow229) or 0
        def _arrow232(projection_1: Callable[[FsCell], int]) -> Callable[[IEnumerable_1[FsCell]], FsCell]:
            def _arrow231(source_2: IEnumerable_1[FsCell]) -> FsCell:
                class ObjectExpr230:
                    @property
                    def Compare(self) -> Callable[[int, int], int]:
                        return compare_primitives

                return max_by(projection_1, source_2, ObjectExpr230())

            return _arrow231

        max_row_index: int = get_index_by(_arrow232) or 0
        return FsColumn(FsRangeAddress(FsAddress(min_row_index, index), FsAddress(max_row_index, index)), cells)

    @property
    def Cells(self, __unit: None=None) -> IEnumerable_1[FsCell]:
        this: FsColumn = self
        return super().Cells(this.cells_004018)

    @property
    def Index(self, __unit: None=None) -> int:
        this: FsColumn = self
        return this.RangeAddress.FirstAddress.ColumnNumber

    @Index.setter
    def Index(self, i: int) -> None:
        this: FsColumn = self
        this.RangeAddress.FirstAddress.ColumnNumber = i or 0
        this.RangeAddress.LastAddress.ColumnNumber = i or 0

    @property
    def MinRowIndex(self, __unit: None=None) -> int:
        this: FsColumn = self
        return this.RangeAddress.FirstAddress.RowNumber

    @property
    def MaxRowIndex(self, __unit: None=None) -> int:
        this: FsColumn = self
        return this.RangeAddress.LastAddress.RowNumber

    def Copy(self, __unit: None=None) -> FsColumn:
        this: FsColumn = self
        ra: FsRangeAddress = this.RangeAddress.Copy()
        def mapping(c: FsCell) -> FsCell:
            return c.Copy()

        cells: IEnumerable_1[FsCell] = map(mapping, this.Cells)
        fcc: FsCellsCollection = FsCellsCollection()
        fcc.AddMany(cells)
        return FsColumn(ra, fcc)

    @staticmethod
    def copy(column: FsColumn) -> FsColumn:
        return column.Copy()

    @staticmethod
    def get_index(column: FsColumn) -> int:
        return column.Index

    def HasCellAt(self, row_index: int) -> bool:
        this: FsColumn = self
        def predicate(c: FsCell) -> bool:
            return c.RowNumber == row_index

        return exists(predicate, this.Cells)

    @staticmethod
    def has_cell_at(row_index: int, column: FsColumn) -> bool:
        return column.HasCellAt(row_index)

    def Item(self, row_index: int) -> FsCell:
        this: FsColumn = self
        return super().Cell(FsAddress(row_index, 1), this.cells_004018)

    @staticmethod
    def item(row_index: int, column: FsColumn) -> FsCell:
        return column.Item(row_index)

    def TryItem(self, row_index: int) -> FsCell | None:
        this: FsColumn = self
        return this.Item(row_index) if this.HasCellAt(row_index) else None

    @staticmethod
    def try_item(row_index: int, column: FsColumn) -> FsCell | None:
        return column.TryItem(row_index)

    def ToDenseColumn(self, __unit: None=None) -> None:
        this: FsColumn = self
        for i in range(this.MinRowIndex, this.MaxRowIndex + 1, 1):
            ignore(this.Item(i))

    @staticmethod
    def to_dense_column(column: FsColumn) -> FsColumn:
        column.ToDenseColumn()
        return column

    @staticmethod
    def create_dense_column_of(column: FsColumn) -> FsColumn:
        new_column: FsColumn = column.Copy()
        new_column.ToDenseColumn()
        return new_column

    def GetEnumerator(self, __unit: None=None) -> IEnumerator[FsCell]:
        this: FsColumn = self
        return get_enumerator(this.Cells)

    def __iter__(self) -> IEnumerator[FsCell]:
        return to_iterator(self.GetEnumerator())

    def System_Collections_IEnumerable_GetEnumerator(self, __unit: None=None) -> IEnumerator[Any]:
        this: FsColumn = self
        return get_enumerator(this)


FsColumn_reflection = _expr233

def FsColumn__ctor_7678C70A(range_address: FsRangeAddress, cells: FsCellsCollection) -> FsColumn:
    return FsColumn(range_address, cells)


__all__ = ["FsColumn_reflection"]




from __future__ import annotations
from typing import Any
from ..fable_library.option import (default_arg, value)
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.seq import (to_array, delay, append, singleton, empty)
from ..fable_library.string_ import replace
from ..fable_library.types import Array
from ..fable_library.util import IEnumerable_1
from .contract import (CLITool, Contract, DTOType, DTO)

def git_with_args(arguments: Array[str]) -> CLITool:
    return CLITool.create("git", arguments)


def create_git_contract_at(path: str, arguments: Array[str]) -> Contract:
    return Contract.create_execute(git_with_args(arguments), path)


def create_git_contract(arguments: Array[str]) -> Contract:
    return Contract.create_execute(git_with_args(arguments))


gitignore_contract: Contract = Contract.create_create(".gitignore", DTOType(10), DTO(1, "# ----- macos rules -----\r\n# taken from https://github.com/github/gitignore/blob/main/Global/macOS.gitignore\r\n# General\r\n.DS_Store\r\n.AppleDouble\r\n.LSOverride\r\n# Icon must end with two \\r\r\nIcon\r\n# Thumbnails\r\n._*\r\n# Files that might appear in the root of a volume\r\n.DocumentRevisions-V100\r\n.fseventsd\r\n.Spotlight-V100\r\n.TemporaryItems\r\n.Trashes\r\n.VolumeIcon.icns\r\n.com.apple.timemachine.donotpresent\r\n# Directories potentially created on remote AFP share\r\n.AppleDB\r\n.AppleDesktop\r\nNetwork Trash Folder\r\nTemporary Items\r\n.apdisk\r\n# ----- windows rules -----\r\n# taken from https://github.com/github/gitignore/blob/main/Global/Windows.gitignore\r\n# Windows thumbnail cache files\r\nThumbs.db\r\nThumbs.db:encryptable\r\nehthumbs.db\r\nehthumbs_vista.db\r\n# Dump file\r\n*.stackdump\r\n# Folder config file\r\n[Dd]esktop.ini\r\n# Recycle Bin used on file shares\r\n$RECYCLE.BIN/\r\n# Windows Installer files\r\n*.cab\r\n*.msi\r\n*.msix\r\n*.msm\r\n*.msp\r\n# Windows shortcuts\r\n*.lnk\r\n# ----- linux rules -----\r\n# taken from https://github.com/github/gitignore/blob/main/Global/Linux.gitignore\r\n*~\r\n# temporary files which can be created if a process still has a handle open of a deleted file\r\n.fuse_hidden*\r\n# KDE directory preferences\r\n.directory\r\n# Linux trash folder which might appear on any partition or disk\r\n.Trash-*\r\n# .nfs files are created when an open file is removed but is still being accessed\r\n.nfs*\r\n"))

gitattributes_contract: Contract = Contract.create_create(".gitattributes", DTOType(10), DTO(1, "**/dataset/** filter=lfs diff=lfs merge=lfs -text"))

def _expr3754() -> TypeInfo:
    return class_type("ARCtrl.Contract.Git.Init", None, Init)


class Init:
    ...

Init_reflection = _expr3754

def _expr3755() -> TypeInfo:
    return class_type("ARCtrl.Contract.Git.Clone", None, Clone)


class Clone:
    ...

Clone_reflection = _expr3755

def Init_get_init(__unit: None=None) -> str:
    return "init"


def Init_get_branchFlag(__unit: None=None) -> str:
    return "-b"


def Init_get_remote(__unit: None=None) -> str:
    return "remote"


def Init_get_add(__unit: None=None) -> str:
    return "add"


def Init_get_origin(__unit: None=None) -> str:
    return "origin"


def Init_createInitContract_6DFDD678(branch: str | None=None) -> Contract:
    branch_1: str = default_arg(branch, "main")
    return create_git_contract([Init_get_init(), Init_get_branchFlag(), branch_1])


def Init_createAddRemoteContract_Z721C83C5(remote_url: str) -> Contract:
    return create_git_contract([Init_get_remote(), Init_get_add(), Init_get_origin(), remote_url])


def Clone_get_clone(__unit: None=None) -> str:
    return "clone"


def Clone_get_branchFlag(__unit: None=None) -> str:
    return "-b"


def Clone_get_noLFSConfig(__unit: None=None) -> str:
    return "-c \"filter.lfs.smudge = git-lfs smudge --skip -- %f\" -c \"filter.lfs.process = git-lfs filter-process --skip\""


def Clone_formatRepoString(username: str, pass_: str, url: str) -> str:
    return replace(url, "https://", "https://" + (((username + ":") + pass_) + "@"))


def Clone_createCloneContract_5000466F(remote_url: str, merge: bool | None=None, branch: str | None=None, token: tuple[str, str] | None=None, nolfs: bool | None=None) -> Contract:
    nolfs_1: bool = default_arg(nolfs, False)
    merge_1: bool = default_arg(merge, False)
    remote_url_1: str = remote_url if (token is None) else Clone_formatRepoString(token[0], token[1], remote_url)
    def _arrow3761(__unit: None=None, remote_url: Any=remote_url, merge: Any=merge, branch: Any=branch, token: Any=token, nolfs: Any=nolfs) -> IEnumerable_1[str]:
        def _arrow3760(__unit: None=None) -> IEnumerable_1[str]:
            def _arrow3759(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow3758(__unit: None=None) -> IEnumerable_1[str]:
                    def _arrow3757(__unit: None=None) -> IEnumerable_1[str]:
                        def _arrow3756(__unit: None=None) -> IEnumerable_1[str]:
                            return singleton(".") if merge_1 else empty()

                        return append(singleton(remote_url_1), delay(_arrow3756))

                    return append(singleton(value(branch)) if (branch is not None) else empty(), delay(_arrow3757))

                return append(singleton(Clone_get_branchFlag()) if (branch is not None) else empty(), delay(_arrow3758))

            return append(singleton(Clone_get_noLFSConfig()) if nolfs_1 else empty(), delay(_arrow3759))

        return append(singleton(Clone_get_clone()), delay(_arrow3760))

    return create_git_contract(to_array(delay(_arrow3761)))


__all__ = ["git_with_args", "create_git_contract_at", "create_git_contract", "gitignore_contract", "gitattributes_contract", "Init_reflection", "Clone_reflection", "Init_get_init", "Init_get_branchFlag", "Init_get_remote", "Init_get_add", "Init_get_origin", "Init_createInitContract_6DFDD678", "Init_createAddRemoteContract_Z721C83C5", "Clone_get_clone", "Clone_get_branchFlag", "Clone_get_noLFSConfig", "Clone_formatRepoString", "Clone_createCloneContract_5000466F"]




from __future__ import annotations
from typing import Any
from ...fable_library.reflection import (TypeInfo, union_type)
from ...fable_library.types import (Array, Union)

def _expr752() -> TypeInfo:
    return union_type("ARCtrl.Process.MaterialType", [], MaterialType, lambda: [[], []])


class MaterialType(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["ExtractName", "LabeledExtractName"]


MaterialType_reflection = _expr752

def MaterialType_create_Z721C83C5(t: str) -> MaterialType:
    if t == "Extract Name":
        return MaterialType(0)

    elif t == "Labeled Extract Name":
        return MaterialType(1)

    else: 
        raise Exception("No other value than \"Extract Name\" or \"Labeled Extract Name\" allowed for materialtype")



def MaterialType__get_AsString(this: MaterialType) -> str:
    if this.tag == 1:
        return "Labeled Extract"

    else: 
        return "Extract"



__all__ = ["MaterialType_reflection", "MaterialType_create_Z721C83C5", "MaterialType__get_AsString"]

